"""
CBMPy: CBRead module
====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBRead.py 669 2019-02-18 22:58:19Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

# __doc__ = """
# CBMPy: CBRead module
# """


import os, time, numpy

# this is a hack that needs to be streamlined a bit
try:
    import cStringIO as csio
except ImportError:
    import io as csio
from . import CBXML, CBModel

from .CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']

##  try:
##  import psyco
##  psyco.full()
##  except:
##  pass

from packaging import version as pkgver

_HAVE_SYMPY_ = False
try:
    import sympy

    if pkgver.parse(sympy.__version__) >= pkgver.Version('0.7.5'):
        _HAVE_SYMPY_ = True

    else:
        del sympy
        print(
            '\nWARNING: SymPy version 0.7.5 or newer is required for symbolic matrix support.'
        )
except ImportError:
    _HAVE_SYMPY_ = False

try:
    import h5py

    _HAVE_HD5_ = True
except ImportError:
    _HAVE_HD5_ = False

_HAVE_XLRD_ = False
try:
    import xlrd

    _HAVE_XLRD_ = True
except ImportError:
    print('\nINFO: No xlrd module available, Excel spreadsheet reading disabled')

__example_models__ = {
    'cbmpy_test_core': 'core_memesa_model.l3.xml',
    'cbmpy_test_ecoli': 'Ecoli_iJR904.glc.l3.xml',
}

## this is now obsolete with the new CBDefaultModels module
# __example_model_path__ = os.path.join(__CBCONFIG__['CBMPY_DIR'], 'models')
# if not os.path.exists(os.path.join(__example_model_path__, 'core_memesa_model.l3.xml')) or\
# not os.path.exists(os.path.join(__example_model_path__, 'Ecoli_iJR904.glc.l3.xml')):
# import zipfile
# print('Installing default models ...')
# zfile = zipfile.ZipFile(os.path.join(__example_model_path__, 'default_models.zip.py'), allowZip64=True)
# zfile.extractall(path=__example_model_path__)
# zfile.close()
# del zipfile, zfile


def loadModel(sbmlfile):
    """
    Loads any SBML model in COBRA, FAME (SBML2FBA), SBML3FBCv1, SBML3FBCv2 format.

     - *sbmlfile* an SBML model file

    """
    mod = res = None
    print('Attempting to load SBML file: {}'.format(sbmlfile))
    try:
        res = CBXML.sbml_fileFindVersion(sbmlfile)[0]
    except Exception as ex:
        print(ex)
        print('\nERROR: cannot determine SBML model type for file {}'.format(sbmlfile))
        return None

    try:
        if res == 'L3V1FBC2':
            mod = readSBML3FBC(sbmlfile)
        elif res == 'L3V1FBC1':
            mod = readSBML3FBC(sbmlfile)
        elif res == 'COBRA':
            mod = readCOBRASBML(sbmlfile, delete_intermediate=True)
        elif res == 'L2FBA':
            mod = readSBML2FBA(sbmlfile)
    except Exception as ex:
        print(ex)
        print('\nERROR: model load failed for SBML file: {}'.format(sbmlfile))
        return None
    return mod


def readSBML3FBC(
    fname,
    work_dir=None,
    return_sbml_model=False,
    xoptions={'validate': False},
    scan_notes_gpr=True,
):
    """
    Read in an SBML Level 3 file with FBC annotation where and return a CBM model object

     - *fname* is the filename
     - *work_dir* is the working directory
     - *return_sbml_model* deprecated and ignored please update code
     - *xoptions* special load options, enable with option=True except for nmatrix_type which has a type.

       - *nogenes* do not load/process genes
       - *noannot* do not load/process any annotations
       - *validate* validate model and display errors and warnings before loading
       - *readcobra* read the cobra annotation
       - *read_model_string* [default=False] read the model from a string (instead of a filename) containing an SBML document
       - *nmatrix_type* [default='normal'] define the type of stoichiometrich matrix to be built

         - 'numpy' dense numpy array (best performance)
         - 'scipy_csr' scipy sparse matrix (lower performance, low memory)
         - 'sympy' a sympy rational matrix (low performance, high memory, cast to dense to analyse)
         - None do not build matrix

    - *scan_notes_gpr* [default=True] if the model is loaded and no genes are detected scan the <notes> field for GPR associationa

    """
    if fname in __example_models__:
        from . import CBDefaultModels

        print(fname)
        xoptions['read_model_string'] = True
        xmod = CBXML.sbml_readSBML3FBC(
            getattr(CBDefaultModels, fname), work_dir, False, xoptions
        )
    else:
        xmod = CBXML.sbml_readSBML3FBC(fname, work_dir, False, xoptions)
    if scan_notes_gpr and len(xmod.getGeneIds()) == 0:
        print(
            'INFO: no standard gene encoding detected, attempting to load from annotations.'
        )
        xmod.createGeneAssociationsFromAnnotations()
    return xmod


def readCOBRASBML(
    fname,
    work_dir=None,
    return_sbml_model=False,
    delete_intermediate=False,
    fake_boundary_species_search=False,
    output_dir=None,
    skip_genes=False,
    scan_notes_gpr=True,
):
    """
    Read in a COBRA format SBML Level 2 file with FBA annotation where and return either a CBM model object
    or a (cbm_mod, sbml_mod) pair if return_sbml_model=True

     - *fname* is the filename
     - *work_dir* is the working directory
     - *delete_intermediate* [default=False] delete the intermediate SBML Level 3 FBC file
     - *fake_boundary_species_search* [default=False] after looking for the boundary_condition of a species search for overloaded id's <id>_b
     - *output_dir* [default=None] the directory to output the intermediate SBML L3 files (if generated) default to input directory
     - *skip_genes* [default=False] do not load GPR data
     - *scan_notes_gpr* [default=True] if the model is loaded and no genes are detected the scan the <notes> field for GPR associationa

    """
    xmod = CBXML.sbml_readCOBRASBML(
        fname,
        work_dir=work_dir,
        return_sbml_model=False,
        delete_intermediate=delete_intermediate,
        fake_boundary_species_search=fake_boundary_species_search,
        output_dir=output_dir,
        skip_genes=skip_genes,
    )
    if scan_notes_gpr and len(xmod.getGeneIds()) == 0:
        xmod.createGeneAssociationsFromAnnotations()
    return xmod


def readSBML2FBA(
    fname,
    work_dir=None,
    return_sbml_model=False,
    fake_boundary_species_search=False,
    scan_notes_gpr=True,
):
    """
    Read in an SBML Level 2 file with FBA annotation where:

     - *fname* is the filename
     - *work_dir* is the working directory if None then only fname is used
     - *return_sbml_model* [default=False] return a a (cbm_mod, sbml_mod) pair
     - *fake_boundary_species_search* [default=False] after looking for the boundary_condition of a species search for overloaded id's <id>_b
     - *scan_notes_gpr* [default=True] if the model is loaded and no genes are detected the scan the <notes> field for GPR associationa

    """

    xmod = CBXML.sbml_readSBML2FBA(
        fname,
        work_dir,
        return_sbml_model=False,
        fake_boundary_species_search=fake_boundary_species_search,
    )
    if scan_notes_gpr and len(xmod.getGeneIds()) == 0:
        xmod.createGeneAssociationsFromAnnotations()
    return xmod


def readLPtoList(fname, work_dir):
    NEW = False
    TYPE = None
    Object = []
    Constr = []
    Bounds = []
    F = file(os.path.join(work_dir, fname), 'r')
    for l in F:
        if l == '' or l[:2] == '\\\\' or l == '\n' or l.strip() == 'END':
            print('skipping')
            if __DEBUG__:
                print(l)
        else:
            L = l.strip()
            if L == 'Maximize':
                TYPE = 'ObjFunc'
                NEW = True
            elif L == 'Subject To':
                TYPE = 'Constr'
                NEW = True
            elif L == 'Bounds':
                TYPE = 'Bounds'
                NEW = True
            else:
                NEW = False

            if TYPE == 'ObjFunc' and not NEW:
                Object.append(L)
            elif TYPE == 'Constr' and not NEW:
                Constr.append(L)
            elif TYPE == 'Bounds' and not NEW:
                Bounds.append(L)

    F.close()
    if __DEBUG__:
        print('ObjectiveLines')
        print(Object)
        print('ConstraintLines')
        print(Constr)
        print('BoundsLines')
        print(Bounds)
    return Object, Constr, Bounds


def readSK_FVA(filename):
    """
    Read Stevens FVA results (opt.fva) file and return a list of dictionaries

    """
    assert os.path.exists(filename), '\nGive me a break!\n'
    ##  name = []
    vari = []
    F = file(filename, 'r')
    for l in F:
        L = l.split(':')
        Jn = L[0].strip()
        Vmin = L[1].strip()
        V = L[2].strip()
        V = V.split('--')
        Vmax = V[0].strip()
        Vstat = V[1].strip()
        ##  name.append((Jn, Vstat))
        ##  vari.append((Vmin, Vmax))
        vari.append({'name': Jn, 'min': Vmin, 'max': Vmax, 'status': Vstat})
        if __DEBUG__:
            print(Jn, Vmin, Vmax, Vstat)
    return vari


##  def readSK_vertexOld(fname, bigfile=False):
##  """
##  Reads in Stevens vertex analysis file and returns:

##  - a list of vertex vectors
##  - a list of ray vectors
##  - the basis of the lineality space as a list of vectors

##  all vectors in terms of the column space of N

##  """

##  assert _HAVE_SYMPY_, 'Install Sympy for rational IO support'

##  assert os.path.exists(fname), 'Uhm exqueeze me ...'
##  SK_vert_file = file(fname, 'r')
##  VertOut = []
##  LinOut = []
##  RayOut = []
##  if bigfile:
##  VertTmp = gzip.open('_vtx_.tmp.gz','wb', compresslevel=3)
##  LinTmp = gzip.open('_lin_.tmp.gz','wb', compresslevel=3)
##  RayTmp = gzip.open('_ray_.tmp.gz','wb', compresslevel=3)
##  GOvert = False
##  GOray = False
##  GOlin = False
##  lcntr = 0
##  lcntrtmp = 0
##  for l in SK_vert_file:
##  lcntr += 1
##  if lcntr == 1000:
##  print 'Processing vertex: %s' % (lcntr + lcntrtmp)
##  lcntrtmp += lcntr
##  lcntr = 0
##  if '* Lineality basis ' in l:
##  GOvert = False
##  GOray = False
##  GOlin = True
##  if '* Rays ' in l:
##  GOvert = False
##  GOray = True
##  GOlin = False
##  if '* Vertices ' in l:
##  GOvert = True
##  GOray = False
##  GOlin = False
##  if l[:2] != '* ':
##  L = l.split()
##  rowL = []
##  for c in L:
##  c = c.strip()
##  if c == '0':
##  rnum = '0'
##  rowL.append(0.0)
##  else:
##  rnum = sympy.Rational('%s' % c)
##  rowL.append(rnum.evalf())
##  del rnum
##  del L
##  if GOlin:
##  if bigfile:
##  rowEnd = len(rowL)
##  cntr = 0
##  for e in rowL:
##  cntr += 1
##  if e == 0.0:
##  LinTmp.write('0.0')
##  else:
##  LinTmp.write('%.14f' % e)
##  if cntr == rowEnd:
##  LinTmp.write('\n')
##  else:
##  LinTmp.write(',')
##  else:
##  LinOut.append(rowL)
##  elif GOray:
##  if bigfile:
##  rowEnd = len(rowL)
##  cntr = 0
##  for e in rowL:
##  cntr += 1
##  if e == 0.0:
##  RayTmp.write('0.0')
##  else:
##  RayTmp.write('%.14f' % e)
##  if cntr == rowEnd:
##  RayTmp.write('\n')
##  else:
##  RayTmp.write(',')
##  else:
##  RayOut.append(rowL)
##  elif GOvert:
##  if bigfile:
##  rowEnd = len(rowL)
##  cntr = 0
##  for e in rowL:
##  cntr += 1
##  if e == 0.0:
##  VertTmp.write('0.0')
##  else:
##  VertTmp.write('%.14f' % e)
##  if cntr == rowEnd:
##  VertTmp.write('\n')
##  else:
##  VertTmp.write(',')
##  else:
##  VertOut.append(rowL)
##  del rowL
##  print '\nProcessed %s vertices.\n' % (lcntr + lcntrtmp)
##  SK_vert_file.close()
##  if bigfile:
##  VertTmp.close()
##  RayTmp.close()
##  LinTmp.close()
##  VertTmp = gzip.open('_vtx_.tmp.gz','rb')
##  LinTmp = gzip.open('_lin_.tmp.gz','rb')
##  RayTmp = gzip.open('_ray_.tmp.gz','rb')
##  return VertTmp, RayTmp, LinTmp
##  else:
##  print 'Lineality basis: %s' % len(LinOut)
##  print 'Number of rays: %s' % len(RayOut)
##  print 'Number of vertices: %s' % len(VertOut)
##  return VertOut, RayOut, LinOut


def readSK_vertexOld(
    fname, bigfile=False, fast_rational=False, nformat='%.14f', compresslevel=3
):
    """
    Reads in Stevens vertex analysis file and returns, even more optimized for large datasets than the original.

     - a list of vertex vectors
     - a list of ray vectors
     - the basis of the lineality space as a list of vectors

    all vectors in terms of the column space of N

    """
    import gzip

    if fast_rational:
        pass
    else:
        assert _HAVE_SYMPY_, 'Install Sympy for rational IO support'

    assert os.path.exists(fname), 'Uhm exqueeze me ...'

    print('\n**********\nreadSK_vertex options are:\n')
    print('bigfile: {}'.format(bigfile))
    print('fast_rational: %s'.format(fast_rational))
    print('nformat: {}'.format(nformat % 0.12345678901234567890))
    print('**********\n')

    SK_vert_file = file(fname, 'r')
    VertOut = []
    LinOut = []
    RayOut = []
    if bigfile:
        VertTmp = gzip.open('_vtx_.tmp.gz', 'wb', compresslevel=compresslevel)
        LinTmp = gzip.open('_lin_.tmp.gz', 'wb', compresslevel=compresslevel)
        RayTmp = gzip.open('_ray_.tmp.gz', 'wb', compresslevel=compresslevel)
    GOvert = False
    GOray = False
    GOlin = False
    lcntr = 0
    lcntrtmp = 0
    TZero = time.time()
    for l in SK_vert_file:
        lcntr += 1
        if lcntr == 1000:
            print(
                'Processing vertex: {} ({} min)'.format(
                    lcntr + lcntrtmp, round((time.time() - TZero) / 60.0, 1)
                )
            )
            lcntrtmp += lcntr
            lcntr = 0
        if '* Lineality basis ' in l:
            GOvert = False
            GOray = False
            GOlin = True
        if '* Rays ' in l:
            GOvert = False
            GOray = True
            GOlin = False
        if '* Vertices ' in l:
            GOvert = True
            GOray = False
            GOlin = False
        if l[:2] != '* ':
            L = l.split()
            rowL = []
            for c in L:
                rnum = None
                c = c.strip()
                if c == '0':
                    rnum = '0'
                    rowL.append(0.0)
                else:
                    if not fast_rational:
                        rnum = sympy.Rational('%s' % c)
                        rowL.append(rnum.evalf())
                        ##  print c, rnum.evalf()
                    else:
                        rnum = c.split('/')
                        if len(rnum) == 1:
                            rowL.append(float(rnum[0]))
                            ##  print c, float(rnum[0])
                        else:
                            rowL.append(float(rnum[0]) / float(rnum[1]))
                            ##  print c, float(rnum[0].strip())/float(rnum[1].strip())
                del rnum
            del L
            rowL = tuple(rowL)
            if GOlin:
                if bigfile:
                    rowEnd = len(rowL)
                    cntr = 0
                    for e in rowL:
                        cntr += 1
                        if e == 0.0:
                            LinTmp.write('0.0')
                        else:
                            LinTmp.write(nformat % e)
                        if cntr == rowEnd:
                            LinTmp.write('\n')
                        else:
                            LinTmp.write(',')
                else:
                    LinOut.append(rowL)
            elif GOray:
                if bigfile:
                    rowEnd = len(rowL)
                    cntr = 0
                    for e in rowL:
                        cntr += 1
                        if e == 0.0:
                            RayTmp.write('0.0')
                        else:
                            RayTmp.write(nformat % e)
                        if cntr == rowEnd:
                            RayTmp.write('\n')
                        else:
                            RayTmp.write(',')
                else:
                    RayOut.append(rowL)
            elif GOvert:
                if bigfile:
                    rowEnd = len(rowL)
                    cntr = 0
                    for e in rowL:
                        cntr += 1
                        if e == 0.0:
                            VertTmp.write('0.0')
                        else:
                            VertTmp.write(nformat % e)
                        if cntr == rowEnd:
                            VertTmp.write('\n')
                        else:
                            VertTmp.write(',')
                else:
                    VertOut.append(rowL)
            del rowL
    print('\nProcessed {} vertices.\n'.format(lcntr + lcntrtmp))
    SK_vert_file.close()
    if bigfile:
        VertTmp.close()
        RayTmp.close()
        LinTmp.close()
        try:
            VertTmp = gzip.open('_vtx_.tmp.gz', 'rb')
            LinTmp = gzip.open('_lin_.tmp.gz', 'rb')
            RayTmp = gzip.open('_ray_.tmp.gz', 'rb')
            return VertTmp, RayTmp, LinTmp
        except Exception as ex:
            print(ex)
            print('\nReturning file names:')
            return '_vtx_.tmp.gz', '_lin_.tmp.gz', '_ray_.tmp.gz'
    else:
        print('Lineality basis: {}'.format(len(LinOut)))
        print('Number of rays: {}'.format(len(RayOut)))
        print('Number of vertices: {}'.format(len(VertOut)))
        return VertOut, RayOut, LinOut


def readSK_vertex(
    fname,
    bigfile=True,
    fast_rational=False,
    nformat='%.14f',
    compression=None,
    hdf5file=None,
):
    """
    Reads in Stevens vertex analysis file:

     - *fname* the input filename (.all file that results from Stevens pipeline)
     - *bigfile* [default=True] this option is now always true and is left in for backwards compatability
     - *fast_rational* [default=False] by default off and uses SymPy for rational-->float conversion, when on uses float decomposition with a slight (2th decimal) decrease in accuracy
     - *nformat* [default='%.14f'] the number format used in output files
     - *compression* [default=None] compression to be used in hdf5 files can be one of [None, 'lzf', 'gz?', 'szip']
     - *hdf5file* [default=None] if None then generic filename '_vtx_.tmp.hdf5' is uses otherwise <hdf5file>.hdf5

    and returns an hdf5 *filename* of the results with a single group named **data** which countains datasets

     - vertices
     - rays
     - lin

    where all vectors are in terms of the column space of N.

    """

    bigfile = True

    if not fast_rational:
        assert _HAVE_SYMPY_, 'Install Sympy for rational IO support'
    if bigfile:
        assert _HAVE_HD5_, 'Install h5py for large dataset support'

    assert os.path.exists(fname), 'Uhm exqueeze me ...'

    print('\n**********\nreadSK_vertex options are:\n')
    print('bigfile: {}'.format(bigfile))
    print('fast_rational: {}'.format(fast_rational))
    print('nformat: {}'.format(nformat % 0.12345678901234567890))
    print('**********\n')

    SK_vert_file = file(fname, 'r')
    VCNTR = 0
    LCNTR = 0
    RCNTR = 0
    CCNTR = 0
    for l in SK_vert_file:
        if '* Vertices (' in l:
            VCNTR = long(l.replace('* Vertices (', '').replace('vectors):', '').strip())
            break
    SK_vert_file.seek(0)
    for l in SK_vert_file:
        if '* Lineality basis (' in l:
            LCNTR = long(
                l.replace('* Lineality basis (', '').replace('vectors):', '').strip()
            )
            break
    SK_vert_file.seek(0)
    for l in SK_vert_file:
        if '* Rays (' in l:
            RCNTR = long(l.replace('* Rays (', '').replace('vectors):', '').strip())
            break
    SK_vert_file.seek(0)
    for l in SK_vert_file:
        if l[:2] != '* ':
            L = l.split()
            CCNTR = len(L)
            del L
            break
    SK_vert_file.seek(0)
    print(RCNTR, LCNTR, VCNTR, CCNTR)
    VertOut = []
    LinOut = []
    RayOut = []
    outFileName = '_vtx_.tmp.hdf5'
    if bigfile:
        if hdf5file != None:
            outFileName = hdf5file + '.' + str(compression) + '.hdf5'
        HD5out = h5py.File(outFileName, 'w')
        if 'data' in HD5out:
            del HD5out['data']
        Dgrp = HD5out.create_group('data')
        if VCNTR > 0:
            VertTmp = HD5out['data'].create_dataset(
                'vertices', (VCNTR, CCNTR), dtype=numpy.double, compression=compression
            )
        if LCNTR > 0:
            LinTmp = HD5out['data'].create_dataset(
                'lin', (LCNTR, CCNTR), dtype=numpy.double, compression=compression
            )
        if RCNTR > 0:
            RayTmp = HD5out['data'].create_dataset(
                'rays', (RCNTR, CCNTR), dtype=numpy.double, compression=compression
            )
    GOvert = False
    GOray = False
    GOlin = False
    lcntr = 0
    lcntrtmp = 0
    TZero = time.time()
    vert_count = 0
    lin_count = 0
    ray_count = 0
    ##  data_row = numpy.zeros((1,CCNTR), 'd')
    print('\nStarting vertex mapping at {}\n'.format(time.strftime('%H:%M:%S')))
    for l in SK_vert_file:
        lcntr += 1
        if lcntr == 1000:
            print(
                'Processing vertex: {}: {} percent @ {} minutes ({} estimated)'.format(
                    lcntr + lcntrtmp,
                    (float(lcntr + lcntrtmp) / float(VCNTR) * 100.0),
                    round((time.time() - TZero) / 60.0, 1),
                    float(VCNTR)
                    / float(lcntr + lcntrtmp)
                    * round((time.time() - TZero) / 60.0, 1),
                )
            )
            lcntrtmp += lcntr
            lcntr = 0
        if '* Lineality basis ' in l and LCNTR > 0:
            GOvert = False
            GOray = False
            GOlin = True
        if '* Rays ' in l and RCNTR > 0:
            GOvert = False
            GOray = True
            GOlin = False
        if '* Vertices ' in l and VCNTR > 0:
            GOvert = True
            GOray = False
            GOlin = False
        if l[:2] != '* ':
            ##  data_row[0,:] = 0.0
            L = l.split()
            if GOlin:
                LinTmp[lin_count] = 0.0
                for c in xrange(CCNTR):
                    rnum = None
                    val = L[c].strip()
                    if val == '0':
                        ##  LinTmp[lin_count,c] = 0.0
                        pass
                    else:
                        if not fast_rational:
                            rnum = sympy.Rational('%s' % val)
                            LinTmp[lin_count, c] = rnum.evalf()
                        else:
                            rnum = val.split('/')
                            if len(rnum) == 1:
                                LinTmp[lin_count, c] = float(rnum[0])
                            else:
                                LinTmp[lin_count, c] = float(rnum[0]) / float(rnum[1])
                lin_count += 1
            elif GOray:
                RayTmp[ray_count] = 0.0
                for c in xrange(CCNTR):
                    rnum = None
                    val = L[c].strip()
                    if val == '0':
                        ##  RayTmp[ray_count,c] = 0.0
                        pass
                    else:
                        if not fast_rational:
                            rnum = sympy.Rational('%s' % val)
                            RayTmp[ray_count, c] = rnum.evalf()
                        else:
                            rnum = val.split('/')
                            if len(rnum) == 1:
                                RayTmp[ray_count, c] = float(rnum[0])
                            else:
                                RayTmp[ray_count, c] = float(rnum[0]) / float(rnum[1])
                ray_count += 1
            elif GOvert:
                VertTmp[vert_count] = 0.0
                for c in xrange(CCNTR):
                    ##  print 'lin_count',lin_count,c
                    ##  print 'ray_count',ray_count,c
                    ##  print 'vert_count',vert_count,c
                    rnum = None
                    try:
                        val = L[c].strip()
                    except Exception as ex:
                        print(ex)
                        print('\nError reading data: your file may be corrupt\n')
                    if val == '0':
                        ##  VertTmp[vert_count,c] = 0.0
                        pass
                    else:
                        if fast_rational:
                            rnum = val.split('/')
                            if len(rnum) == 1:
                                VertTmp[vert_count, c] = float(rnum[0])
                            else:
                                VertTmp[vert_count, c] = float(rnum[0]) / float(rnum[1])
                        else:
                            rnum = sympy.Rational('%s' % val)
                            VertTmp[vert_count, c] = rnum.evalf()

                vert_count += 1

    print('\nProcessed {} vectors.\n'.format(lcntr + lcntrtmp))
    SK_vert_file.close()
    if bigfile:
        HD5out.close()
        ##  HD5out = h5py.File(outFileName,'r')
        return outFileName
    else:
        print('Lineality basis: {}'.format(len(LinOut)))
        print('Number of rays: {}'.format(len(RayOut)))
        print('Number of vertices: {}'.format(len(VertOut)))
        return VertOut, RayOut, LinOut


def readExcel97Model(xlname, write_sbml=True, sbml_level=3, return_dictionaries=False):
    """
    Reads a model encoded as an Excel97 workbook and returns it as a CBMPy model object and SBML file. Note the workbook must be formatted
    exactly like those produced by cbm.writeModelToExcel97(). Note that reactions have to be defined in **both** the *reaction*
    and *network_react* sheets to be included in the model.

     - *xlpath* the filename of the Excel workbook
     - *return_model* [default=True] construct and return the CBMPy model
     - *write_sbml* [default=True] write the SBML file to fname
     - *return_dictionaries* [default=False] return the dictionaries constructed when reading the Excel file (in place of the model)
     - *sbml_level* [default=3] write the SBML file as either SBML L2 FBA or SBML L3 FBC file.


    """

    if not _HAVE_XLRD_:
        print(
            '\nERROR: Cannot read Excel file, XLRD package not available (http://pypi.python.org/pypi/xlrd)'
        )
        return

    assert os.path.exists(xlname), '\nERROR: File "{}" does not exist'.format(xlpath)

    def c2s(c):
        """
        Utility function converting a XLRD cell to a string
        """
        # return str(c.value.strip())
        return str(c.value)

    MSGLog = csio.StringIO()

    def logMsg(msg):
        """
        Message logging utility

        """
        print(msg)
        MSGLog.write('{}\n'.format(msg))

    wb = xlrd.open_workbook(xlname)

    for s in wb.sheets():
        logMsg('Sheet: {}'.format(s.name))

    sI = wb.sheet_by_name('info')
    sM = wb.sheet_by_name('metabolites')
    sR = wb.sheet_by_name('reactions')
    sNR = wb.sheet_by_name('network_react')
    sNM = wb.sheet_by_name('network_metab')
    sS = wb.sheet_by_name('solution')
    sMI = wb.sheet_by_name('miriam')
    try:
        sComp = wb.sheet_by_name('compartments')
    except xlrd.XLRDError:
        sComp = None

    # INFO
    dInfo = {}
    for r_ in range(3):
        r = sI.row(r_)
        if r_ == 2:
            obj = {}
            obj['osense'] = c2s(r.pop(1))
            coeff = True
            cout = []
            for c in range(1, len(r)):
                if r[c].value != '' and c2s(r[c]) != '':
                    if coeff:
                        cout.append([r[c].value])
                        coeff = False
                    else:
                        cout[-1].append(c2s(r[c]))
                        coeff = True
            obj['objflux'] = cout
            dInfo['objective'] = obj
            del cout, obj
        else:
            dInfo[c2s(r[0])] = c2s(r[1])

    # metabolites
    mcolNames = [c2s(c) for c in sM.row(0)]

    dCompartments = {}

    dMet = {}
    for r_ in range(1, sM.nrows):
        r = sM.row(r_)
        met = {}
        annot = {}
        for c_ in range(1, len(mcolNames)):
            if c_ < 6:
                if mcolNames[c_] == 'charge':
                    met[mcolNames[c_]] = r[c_].value
                elif mcolNames[c_] == 'fixed':
                    met[mcolNames[c_]] = bool(r[c_].value)
                else:
                    met[mcolNames[c_]] = c2s(r[c_])
                if mcolNames[c_] == 'compartment':
                    dCompartments[c2s(r[c_])] = {'id': c2s(r[c_])}
            else:
                annot[mcolNames[c_]] = c2s(r[c_])
        met['annot'] = annot
        met['id'] = c2s(r[0])
        dMet[c2s(r[0])] = met

    # reactions
    rcolNames = [c2s(c) for c in sR.row(0)]

    dReact = {}
    for r_ in range(1, sR.nrows):
        r = sR.row(r_)
        reac = {}
        annot = {}
        for c_ in range(1, len(rcolNames)):
            if c_ < 6:
                if rcolNames[c_] == 'lowerbound' or rcolNames[c_] == 'upperbound':
                    reac[rcolNames[c_]] = r[c_].value
                elif rcolNames[c_] == 'reversible':
                    rev = None
                    if r[c_].ctype == 1:
                        rev = c2s(r[c_])
                        if rev in ['True', 'Yes', True, 'yes', 'true']:
                            rev = True
                        elif rev in ['False', 'No', False, 'no', 'True']:
                            rev = False
                    else:
                        rev = bool(r[c_].value)
                    reac[rcolNames[c_]] = rev
                else:
                    reac[rcolNames[c_]] = c2s(r[c_])
                if rcolNames[c_] == 'compartment':
                    dCompartments[c2s(r[c_])] = {'id': c2s(r[c_])}
            else:
                annot[rcolNames[c_]] = c2s(r[c_])
        reac['annot'] = annot
        reac['id'] = c2s(r[0])
        dReact[c2s(r[0])] = reac

    # reaction network
    nDict = {}
    reagentList = []
    for n_ in range(0, sNR.nrows, 3):
        coeff = True
        sub = []
        s = sNR.row(n_ + 1)
        for c in range(1, len(s)):
            if s[c].value != '' and c2s(s[c]) != '':
                if coeff:
                    sub.append([s[c].value])
                    coeff = False
                else:
                    sub[-1].append(c2s(s[c]))
                    coeff = True
                    if c2s(s[c]) not in reagentList:
                        reagentList.append(c2s(s[c]))

        coeff = True
        prod = []
        p = sNR.row(n_ + 2)
        for c in range(1, len(p)):
            if p[c].value != '' and c2s(p[c]) != '':
                if coeff:
                    prod.append([p[c].value])
                    coeff = False
                else:
                    prod[-1].append(c2s(p[c]))
                    coeff = True
                    if c2s(p[c]) not in reagentList:
                        reagentList.append(c2s(p[c]))
        nDict[c2s(sNR.row(n_)[0])] = {'subs': sub, 'prod': prod}

    # compartments
    if sComp != None:

        # metabolites
        ccolNames = [c2s(c) for c in sComp.row(0)]

        for m_ in range(1, sComp.nrows):
            r = sComp.row(m_)
            comp = {}
            annot = {}
            for c_ in range(1, len(ccolNames)):
                if c_ < 6:
                    if ccolNames[c_] == 'name':
                        comp[ccolNames[c_]] = c2s(r[c_])
                    elif ccolNames[c_] == 'size':
                        comp[ccolNames[c_]] = r[c_].value
                    elif ccolNames[c_] == 'dimensions':
                        comp[ccolNames[c_]] = r[c_].value
                    elif ccolNames[c_] == '# species' or ccolNames[c_] == '# reactions':
                        pass
                else:
                    annot[ccolNames[c_]] = c2s(r[c_])
            comp['annot'] = annot
            comp['id'] = c2s(r[0])
            dCompartments[c2s(r[0])] = comp

    # MIRIAM
    mirNames = [c2s(c) for c in sMI.row(0)]
    dMir = {}
    for m_ in range(1, sMI.nrows):
        r = sMI.row(m_)
        qualD = {}
        for c_ in range(1, len(mirNames)):
            if c2s(r[c_]) != '':
                if mirNames[c_] in qualD:
                    qualD[mirNames[c_]].append(c2s(r[c_]))
                else:
                    qualD[mirNames[c_]] = [c2s(r[c_])]
        dMir[c2s(r[0])] = qualD

    # clean up
    if '' in dCompartments:
        dCompartments.pop('')
    if '' in nDict:
        nDict.pop('')
    if '' in dReact:
        dReact.pop('')
    if '' in dMet:
        dMet.pop('')

    # cross check reactions
    dReactError = {}
    nDictError = {}
    for r_ in tuple(dReact):
        if r_ not in nDict:
            logMsg(
                'ERROR: Reaction "{}" listed but not defined ... removing.'.format(r_)
            )
            dReactError[r_] = dReact.pop(r_)
    for r_ in tuple(nDict):
        if r_ not in dReact:
            logMsg(
                'ERROR: Reaction "{}" defined but not listed ... removing.'.format(r_)
            )
            nDictError[r_] = nDict.pop(r_)

    # cross check reagents versus species
    nDictReagentError = {}
    reagentError = []

    for s_ in dMet:
        if s_ not in reagentList:
            logMsg('WARNING: Metabolite "{}" listed but not used.'.format(s_))
            # dMetError[s_] = dMet.pop(s_)

    for s_ in reagentList:
        if s_ not in dMet:
            logMsg('ERROR: Reagent "{}" used but not defined.'.format(s_))
            reagentError.append(reagentList.pop(reagentList.index(s_)))

    # remove reactions that contain unknown reagents (undefined species)
    if len(reagentError) > 0:
        for e_ in tuple(nDict):
            reag = [a[1] for a in nDict[e_]['subs']] + [a[1] for a in nDict[e_]['prod']]
            for r_ in reag:
                if r_ in reagentError:
                    logMsg(
                        'ERROR: Reaction "{}" contains an unknown reagent "{}" ... removing'.format(
                            e_, r_
                        )
                    )
                    nDictReagentError[e_] = {
                        'reaction': dReact.pop(e_),
                        'stoich': nDict.pop(e_),
                    }
    if return_dictionaries:
        return (dInfo, dReact, dMet, nDict, dCompartments, dMir)

    # assemble CBMPy model

    # construct model object
    cmod = CBModel.Model(dInfo['id'])
    cmod.name = dInfo['name']
    cmod.description = 'Model created from Excel spreadsheet: {}'.format(xlname)
    cmod.sourcefile = xlname
    cmod.setCreatedDate()

    # add compartments
    for c_ in dCompartments:
        CC = dCompartments[c_]
        if 'name' in CC:
            name = CC['name']
        else:
            name = CC['id']
        if 'size' in CC:
            size = CC['size']
        else:
            size = 1.0
        if 'dimensions' in CC:
            dimensions = CC['dimensions']
        else:
            dimensions = 3
        C = CBModel.Compartment(CC['id'], name, size, dimensions)
        if 'annot' in CC:
            for a_ in CC['annot']:
                C.setAnnotation(a_, CC['annot'][a_])
        cmod.addCompartment(C)
    del c_, CC, name, size, dimensions

    # add metabolites
    for m_ in dMet:
        M = dMet[m_]
        S = CBModel.Species(
            M['id'],
            boundary=M['fixed'],
            name=M['name'],
            value=0.0,
            compartment=M['compartment'],
            charge=M['charge'],
            chemFormula=M['chemformula'],
        )
        for a_ in M['annot']:
            S.setAnnotation(a_, M['annot'][a_])
        cmod.addSpecies(S)
    del m_, M, S

    # add reactions
    for r_ in nDict:
        Rn = nDict[r_]
        Ri = dReact[r_]
        R = CBModel.Reaction(Ri['id'], Ri['name'], reversible=Ri['reversible'])
        for s_ in Rn['subs']:
            R.createReagent(s_[1], -s_[0])
        for p_ in Rn['prod']:
            R.createReagent(p_[1], p_[0])
        for a_ in Ri['annot']:
            R.setAnnotation(a_, Ri['annot'][a_])
        cmod.addReaction(R, create_default_bounds=False)
    # cross reference reagents (shortcut)
    for s_ in cmod.species:
        tmp = s_.isReagentOf()
    del r_, s_, p_, R, Rn, Ri, tmp

    # add flux bounds
    for r_ in dReact:
        lb = dReact[r_]['lowerbound']
        ub = dReact[r_]['upperbound']
        try:
            lb = float(lb)
        except:
            if dReact[r_]['reversible']:
                lb = -numpy.inf
            else:
                lb = 0.0
            logMsg(
                'Undefined lower bound for reaction "{}" setting to {}'.format(r_, lb)
            )
        try:
            ub = float(ub)
        except:
            ub = numpy.inf
            logMsg(
                'Undefined upper bound for reaction "{}" setting to {}'.format(r_, ub)
            )
        cmod.createReactionLowerBound(r_, lb)
        cmod.createReactionUpperBound(r_, ub)
    del r_, lb, ub

    # add objective function
    O = CBModel.Objective('obj1', dInfo['objective']['osense'])
    O.createFluxObjectives(dInfo['objective']['objflux'])
    cmod.addObjective(O, True)
    del O

    # markup with MIRIAM
    for x_ in cmod.species + cmod.reactions + cmod.compartments:
        xid = x_.getId()
        if xid in dMir:
            for q_ in dMir[xid]:
                for i_ in dMir[xid][q_]:
                    # ent, mid = i_.rsplit('/',1)
                    # x_.addMIRIAMannotation(q_, ent, mid)
                    # if q_ == 'is':
                    # qual = 'isA'
                    # else:
                    # qual = q_
                    if x_.miriam == None:
                        x_.miriam = CBModel.MIRIAMannotation()
                    x_.miriam.addIDorgURI(q_, i_)

    # ok lets play
    # try:
    # cmod.createGeneAssociationsFromAnnotations()
    # geneerrors = cmod.testGeneProteinAssociations()
    # logMsg('Successfully created gene associations from annotations.')
    # except Exception as ex:
    # logMsg(ex)
    # try:
    # cbm.analyzeModel(cmod)
    # logMsg('Successfully optimized model.')
    # except Exception as ex:
    # logMsg(ex)
    # try:
    # cbm.writeModelToExcel97(cmod, xlname.replace('.xls','')+'.new')
    # logMsg('Successfully wrote model to new Excel spreadsheet: "{}"'.format(xlname.replace('.xls','')+'.DEBUG.new.xls'))
    # except Exception as ex:
    # logMsg(ex)
    if write_sbml:
        try:
            if sbml_level == 3:
                CBXML.sbml_writeSBML3FBC(cmod, xlname + '.l3.xml', gpr_from_annot=False)
                logMsg(
                    'Successfully wrote model SBML3FBC file: "{}"'.format(
                        xlname + '.l3.xml'
                    )
                )
            else:
                CBXML.sbml_writeSBML2FBA(cmod, xlname + '.l3.xml')
                logMsg(
                    'Successfully wrote model SBML2FBA file: "{}"'.format(
                        xlname + '.l3.xml'
                    )
                )
        except Exception as ex:
            logMsg(ex)

    # write log to file
    """
    F = file(os.path.join(cDir,'{}.log'.format(xlname)),'w')
    MSGLog.seek(0)
    F.write(MSGLog.read())
    F.close()
    """
    print('\n*****\nExcel97 Read Log\n*****\n')
    MSGLog.seek(0)
    print(MSGLog.read())
    print('')
    MSGLog.close()

    return cmod



'''
This code is adapted from https://gist.github.com/adewes/5884820

Thanks to adewes for putting it online

TODO: clarify copyright - if I would not be so lazy I could have come up with it
myself
'''

import random

myrandom = random.Random(135)


def get_random_color():
    x = myrandom.uniform(0, 1.0)
    y = myrandom.uniform(0.7, 1.0)
    z = myrandom.uniform(0.7, 1.0)
    return (x, y, z)


def color_distance(c1, c2):
    return min(abs(c1[0] - c2[0]), 1 - abs(c1[0] - c2[0])) + abs(c1[1] - c2[1]) + abs(c1[2] - c2[2])


def generate_new_color(existing_colors):
    """ creates a new random color (in hsv scheme) that is most different
    to the existing colors.
    """
    max_distance = None
    best_color = None
    for i in range(0, 100):
        color = get_random_color()
        if not existing_colors:
            return color
        best_distance = min([color_distance(color, c) for c in existing_colors])
        if not max_distance or best_distance > max_distance:
            max_distance = best_distance
            best_color = color
    return best_color




"""
CBMPy: CBConfig module
======================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBConfig.py 711 2020-04-27 14:22:34Z bgoli $)

"""
# gets rid of "invalid variable name" info
# pylint: disable=C0103
# gets rid of "line to long" info
# pylint: disable=C0301
# use with caution: gets rid of module xxx has no member errors (run once enabled)
# pylint: disable=E1101

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import platform

__VERSION_MAJOR__ = 0
__VERSION_MINOR__ = 8
__VERSION_MICRO__ = 2

__CBCONFIG__ = {
    'VERSION_MAJOR': __VERSION_MAJOR__,
    'VERSION_MINOR': __VERSION_MINOR__,
    'VERSION_MICRO': __VERSION_MICRO__,
    'VERSION_STATUS': '',
    'VERSION': '{}.{}.{}'.format(
        __VERSION_MAJOR__, __VERSION_MINOR__, __VERSION_MICRO__
    ),
    'DEBUG': False,
    'SOLVER_PREF': 'CPLEX',
    #'SOLVER_PREF': 'GLPK',
    'SOLVER_ACTIVE': None,
    'REVERSIBLE_SYMBOL': '<==>',
    'IRREVERSIBLE_SYMBOL': '-->',
    'HAVE_SBML2': False,
    'HAVE_SBML3': False,
    'CBMPY_DIR': None,
    'SYMPY_DENOM_LIMIT': 10 ** 32,
    'ENVIRONMENT': '{} {} ({})'.format(
        platform.system(), platform.release(), platform.architecture()[0]
    ),
}


def current_version():
    """
    Return the current CBMPy version as a string

    """
    return '{}.{}.{}'.format(__VERSION_MAJOR__, __VERSION_MINOR__, __VERSION_MICRO__)


def current_version_tuple():
    """
    Return the current CBMPy version as a tuple (x, y, z)

    """
    return (__VERSION_MAJOR__, __VERSION_MINOR__, __VERSION_MICRO__)



"""
CBMPy: CBQt4 module
===================
Constraint Based Modelling in Python (http://pysces.sourceforge.net/getNewReaction)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBQt4.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import os
import time
import random
import math
import re
import webbrowser
import urllib2
from .CBCommon import pp_chemicalFormula
from .CBModel import Reaction as CBMReaction


HAVE_QT4 = False
try:
    import PyQt4
    from PyQt4 import QtCore, QtGui, QtSvg
    from PyQt4.QtWebKit import QGraphicsWebView

    HAVE_QT4 = True
    print('Qt4 GUI tools available')
except ImportError as ex:
    print('\nQt4 GUI tools not available.')
    print(ex)


class ReactionCreator(QtGui.QWidget):

    # _fba = None
    _mlist = None
    _rlist = None
    _flist = None
    _newSubs = None
    _newProds = None
    _fixColour = None
    _errColour = None
    _goodColour = None
    _cfdict = None
    _cndict = None
    _ccdict = None
    IGNORECHECK = False
    ISBALANCED = True
    NewReaction = None
    _Blower = '-inf'
    _Bupper = 'inf'

    def __init__(self, rlist, mlist, flist, cfdict, cndict, ccdict):
        super(ReactionCreator, self).__init__()
        self.mousePos = self.cursor().pos()
        rlist.sort()
        mlist.sort()
        flist.sort()
        self._rlist = rlist
        self._mlist = mlist
        self._flist = flist
        self._cfdict = cfdict
        self._cndict = cndict
        self._ccdict = ccdict
        self._newSubs = []
        self._newProds = []
        self._fixColour = QtGui.QColor(0, 0, 153, alpha=255)
        self._errColour = QtGui.QColor(255, 0, 0, alpha=255)
        self._goodColour = QtGui.QColor(0, 100, 0, alpha=255)
        self.initUI()

    def addSubstrate(self, coeff, sid):
        self.tblSub.insertRow(self.tblSubRow)
        self.tblSub.setItem(
            self.tblSubRow, 0, QtGui.QTableWidgetItem('{}'.format(coeff))
        )
        self.tblSub.setItem(self.tblSubRow, 1, QtGui.QTableWidgetItem('{}'.format(sid)))
        CF = 'None'
        if sid in self._cfdict:
            CF = self._cfdict[sid]
        self.tblSub.item(self.tblSubRow, 1).setToolTip(CF)
        if sid in self._flist:
            self.tblSub.item(self.tblSubRow, 1).setForeground(self._fixColour)
        self.tblSub.item(self.tblSubRow, 0).setTextAlignment(QtCore.Qt.AlignCenter)
        self.tblSubRow += 1

    def addSelectedSubstrates(self):
        self.IGNORECHECK = True
        items = [str(it_.text()) for it_ in self.lstSub.selectedItems()]
        self.setFocus(PyQt4.QtCore.Qt.OtherFocusReason)
        # print(items)
        for i_ in items:
            if i_ not in self._newSubs:
                self.addSubstrate(1, i_)
                self._newSubs.append(i_)
        self.IGNORECHECK = False
        self.statusBar.showMessage('Substrates(s) added')
        self.checkBalance()

    def addProduct(self, coeff, sid):
        self.tblProd.insertRow(self.tblProdRow)
        self.tblProd.setItem(
            self.tblProdRow, 0, QtGui.QTableWidgetItem('{}'.format(coeff))
        )
        self.tblProd.setItem(
            self.tblProdRow, 1, QtGui.QTableWidgetItem('{}'.format(sid))
        )
        CF = 'None'
        if sid in self._cfdict:
            CF = self._cfdict[sid]
        self.tblProd.item(self.tblProdRow, 1).setToolTip(CF)
        if sid in self._flist:
            self.tblProd.item(self.tblProdRow, 1).setForeground(self._fixColour)
        self.tblProd.item(self.tblProdRow, 0).setTextAlignment(QtCore.Qt.AlignCenter)
        self.tblProdRow += 1

    def addSelectedProducts(self):
        self.IGNORECHECK = True
        items = [str(it_.text()) for it_ in self.lstProd.selectedItems()]
        self.setFocus(PyQt4.QtCore.Qt.OtherFocusReason)
        # print(items)
        for i_ in items:
            if i_ not in self._newProds:
                self.addProduct(1, i_)
                self._newProds.append(i_)
        self.IGNORECHECK = False
        self.statusBar.showMessage('Product(s) added')
        self.checkBalance()

    # def keyPressEvent(self, event):
    # print('KeyPress key: {}'.format(str(event.key())))
    # if event.key() == 16777223:
    # print('You pressed the delete key')

    def deleteSubstrates(self):
        self.deleteReagents('substrate')

    def deleteProducts(self):
        self.deleteReagents('product')

    def deleteAllSubstrates(self):
        self.tblSub.clear()
        for r_ in range(self.tblSubRow - 1, -1, -1):
            self.tblSub.removeRow(r_)
        self.tblSubRow = 0
        self.checkBalance()

    def deleteAllProducts(self):
        self.tblProd.clear()
        for r_ in range(self.tblProdRow - 1, -1, -1):
            self.tblProd.removeRow(r_)
        self.tblProdRow = 0
        self.checkBalance()

    def deleteReagents(self, reagentType):
        selected = None
        PRODACTIVE = False
        SUBACTIVE = False
        self.IGNORECHECK = True
        if reagentType == 'substrate':
            selected = [
                (it_.row(), it_.column()) for it_ in self.tblSub.selectedItems()
            ]
            SUBACTIVE = True
        elif reagentType == 'product':
            selected = [
                (it_.row(), it_.column()) for it_ in self.tblProd.selectedItems()
            ]
            PRODACTIVE = True

        if selected != None:
            deleteRow = []
            if len(selected) == 2:
                if selected[0][0] == selected[1][0]:
                    if selected[0][1] + 1 == selected[1][1]:
                        deleteRow.append(selected[0][0])
            elif len(selected) > 2:
                for it_ in range(0, len(selected), 2):
                    if selected[it_][1] == selected[it_ + 1][1]:
                        if selected[it_][0] + 1 == selected[it_ + 1][0]:
                            if selected[it_][0] not in deleteRow:
                                deleteRow.append(selected[it_][0])
                            if selected[it_][0] + 1 not in deleteRow:
                                deleteRow.append(selected[it_][0] + 1)
            deleteRow.sort()
            for d_ in range(len(deleteRow) - 1, -1, -1):
                if SUBACTIVE:
                    print('Deleting Sub table row: {}'.format(deleteRow[d_]))
                    self.statusBar.showMessage('Substrate(s) deleted')
                    # print(self._newSubs)
                    # print(str(self.tblSub.item(d_, 1).text()))
                    self._newSubs.pop(
                        self._newSubs.index(str(self.tblSub.item(d_, 1).text()))
                    )
                    self.tblSub.removeRow(deleteRow[d_])
                    self.tblSubRow -= 1
                elif PRODACTIVE:
                    print('Deleting Prod table row: {}'.format(deleteRow[d_]))
                    self.statusBar.showMessage('Product(s) deleted')
                    # print(self._newProds)
                    # print(str(self.tblProd.item(d_, 1).text()))
                    self._newProds.pop(
                        self._newProds.index(str(self.tblProd.item(d_, 1).text()))
                    )
                    self.tblProd.removeRow(deleteRow[d_])
                    self.tblProdRow -= 1
        self.IGNORECHECK = False
        self.checkBalance()

    def checkBalance(self):
        if self.IGNORECHECK:
            return
        output = {}
        left = {}
        right = {}
        for r_ in range(self.tblSubRow):
            sid = str(self.tblSub.item(r_, 1).text())
            scoef = float(str(self.tblSub.item(r_, 0).text()))
            # print scoef
            if sid in self._cfdict:
                cf = self._cfdict[sid]
                if cf not in [None, 'None', '', ' ']:
                    cfl = pp_chemicalFormula.parseString(cf).asList()
                else:
                    cfl = []
                # print sid, cf, cfl
                for e_ in cfl:
                    if e_[0] in output:
                        output[e_[0]] = output[e_[0]] + -scoef * float(e_[1])
                        # print scoef*float(e_[1])
                    else:
                        output[e_[0]] = -scoef * float(e_[1])
                        # print scoef*float(e_[1])
                    if e_[0] in left:
                        left[e_[0]] = left[e_[0]] + scoef * float(e_[1])
                    else:
                        left[e_[0]] = scoef * float(e_[1])
        for r_ in range(self.tblProdRow):
            sid = str(self.tblProd.item(r_, 1).text())
            pcoef = float(str(self.tblProd.item(r_, 0).text()))
            # print pcoef, type(pcoef)
            if sid in self._cfdict:
                cf = self._cfdict[sid]
                if cf not in [None, 'None', '', ' ']:
                    cfl = pp_chemicalFormula.parseString(cf).asList()
                else:
                    cfl = []
                # print sid, cf, cfl
                for e_ in cfl:
                    if e_[0] in output:
                        # print -pcoef*float(e_[1])
                        output[e_[0]] = output[e_[0]] + pcoef * float(e_[1])
                    else:
                        # print -pcoef*float(e_[1])
                        output[e_[0]] = pcoef * float(e_[1])
                    if e_[0] in right:
                        # print -pcoef*float(e_[1])
                        right[e_[0]] = right[e_[0]] + pcoef * float(e_[1])
                    else:
                        # print -pcoef*float(e_[1])
                        right[e_[0]] = pcoef * float(e_[1])
        # print output
        self.updateBalance(output, left, right)
        # self.txtBal.setText(str(output))

    def updateBalance(self, bdict, left, right):
        colHead = []
        keys = list(bdict)
        if self.tblBalCol > 0:
            for c_ in range(self.tblBalCol - 1, -1, -1):
                self.tblBal.removeColumn(c_)
        self.tblBalCol = 0
        self.ISBALANCED = True
        for k_ in range(len(keys)):
            if not keys[k_] in left:
                left[keys[k_]] = 'None'
            if not keys[k_] in right:
                right[keys[k_]] = 'None'
            self.tblBalCol += 1
            self.tblBal.insertColumn(k_)
            self.tblBal.setItem(
                0, k_, QtGui.QTableWidgetItem('{}'.format(abs(bdict[keys[k_]])))
            )
            self.tblBal.setItem(
                1, k_, QtGui.QTableWidgetItem('{}'.format(left[keys[k_]]))
            )
            self.tblBal.setItem(
                2, k_, QtGui.QTableWidgetItem('{}'.format(right[keys[k_]]))
            )
            self.tblBal.item(0, k_).setTextAlignment(QtCore.Qt.AlignCenter)
            if k_ == 0:
                boldFont = self.tblBal.item(0, k_).font()
                boldFont.setBold(True)
            self.tblBal.item(0, k_).setFont(boldFont)
            self.tblBal.item(1, k_).setTextAlignment(QtCore.Qt.AlignCenter)
            self.tblBal.item(2, k_).setTextAlignment(QtCore.Qt.AlignCenter)
            if bdict[keys[k_]] != 0.0:
                self.ISBALANCED = False
                self.tblBal.item(0, k_).setForeground(self._errColour)
            else:
                self.tblBal.item(0, k_).setForeground(self._goodColour)
        self.tblBal.setHorizontalHeaderLabels(QtCore.QStringList(keys))

    def showErrorMessage(self, errorMsg, title="Reaction Creator"):
        QtGui.QMessageBox.critical(
            None,
            title,
            errorMsg,
            QtGui.QMessageBox.Ok | QtGui.QMessageBox.Default,
            QtGui.QMessageBox.NoButton,
        )
        self.statusBar.showMessage(errorMsg)

    def getNewReaction(self):
        reversible = self.btReverse.isChecked()
        if not reversible and self._Blower == '-inf':
            self._Blower = 0.0
        Id = str(self.txtId.toPlainText()).strip()
        Name = str(self.txtNm.toPlainText()).strip()

        if Id == 'NewReactionId':
            print('\nWARNING: using default reaction id')
        errorMsg = None
        if Id == "":
            errorMsg = 'Reaction ID must be specified.'
        elif Id in self._rlist:
            errorMsg = 'Reaction ID \"{}\" already exists.'.format(Id)
        if errorMsg != None:
            self.showErrorMessage(errorMsg)
            self.NewReaction = None
            return None
        if self.tblSubRow == 0 and self.tblProdRow == 0:
            self.showErrorMessage('At least one reagent must be defined.')
            self.NewReaction = None
            return None

        # print('\nid=\"{}\"\nname=\"{}\"'.format(Id, Name))
        # print self.tblSub.rowCount(), self.tblSubRow
        # print self.tblProd.rowCount(), self.tblProdRow
        Reag = {}
        exReac = False
        for s_ in range(self.tblSub.rowCount()):
            coeff = -abs(float(self.tblSub.item(s_, 0).text()))
            Sid = str(self.tblSub.item(s_, 1).text()).strip()
            if Sid in self._flist:
                exReac = True
            Reag[Sid] = coeff
        for p_ in range(self.tblProd.rowCount()):
            coeff = abs(float(self.tblProd.item(p_, 0).text()))
            Sid = str(self.tblProd.item(p_, 1).text()).strip()
            if Sid in self._flist:
                exReac = True
            if Sid in Reag:
                Reag[Sid] += coeff
            else:
                Reag[Sid] = coeff

        for r_ in tuple(Reag):
            if Reag[r_] == 0.0:
                Reag.pop(r_)
                print('removing zero coefficient reagent: {}'.format(r_))
        self.NewReaction = {
            'reversible': reversible,
            'id': Id,
            'name': Name,
            'is_exchange': exReac,
            'is_balanced': self.ISBALANCED,
            'reagents': Reag,
            'upper_bound': self._Bupper,
            'lower_bound': self._Blower,
        }

        sub = ''
        prod = ''
        for r_ in Reag:
            coeff = abs(Reag[r_])
            if Reag[r_] < 0.0:
                if coeff == 1.0:
                    sub += '%s + ' % (r_)
                else:
                    sub += '{%s} %s + ' % (coeff, r_)
            else:
                if coeff == 1.0:
                    prod += '%s + ' % (r_)
                else:
                    prod += '{%s} %s + ' % (coeff, r_)
        if reversible:
            eq = '%s\n\t%s\n%s' % (sub[:-3], '<==>', prod[:-2])
        else:
            eq = '%s\n\t%s\n%s' % (sub[:-3], '-->', prod[:-2])

        # quit_msg = "Add reaction:\n\n{}\n\t{}\n{}".format(sub,rev,prod)
        quit_msg = eq
        reply = QtGui.QMessageBox.question(
            self,
            'Do you want to add the reaction \"{}\" to the model?'.format(Id),
            quit_msg,
            QtGui.QMessageBox.Yes,
            QtGui.QMessageBox.No,
        )
        if reply == QtGui.QMessageBox.Yes:
            self.statusBar.showMessage('Reaction {} added to model'.format(Id))
            print(self.NewReaction)
            QtGui.qApp.quit()
        else:
            print('Try again')
            self.NewReaction = None

    def initUI(self):
        # create labels
        lblSub = QtGui.QLabel('Substrates')
        lblSub.setAlignment(QtCore.Qt.AlignCenter)
        lblProd = QtGui.QLabel('Products')
        lblProd.setAlignment(QtCore.Qt.AlignCenter)
        lblId = QtGui.QLabel('Reaction id')
        lblId.setAlignment(QtCore.Qt.AlignCenter)
        lblNm = QtGui.QLabel('Reaction name')
        lblNm.setAlignment(QtCore.Qt.AlignCenter)

        # create text boxes
        self.txtId = QtGui.QTextEdit()
        self.txtId.setMaximumHeight(25)
        self.txtId.setText('NewReactionId')
        self.txtNm = QtGui.QTextEdit()
        self.txtNm.setMaximumHeight(25)
        self.txtNm.setText('NewReactionName')
        self.txtBal = QtGui.QTextEdit()
        self.txtBal.setMaximumHeight(40)

        # create static lists
        self.lstSub = QtGui.QListWidget()
        self.lstSub.setSelectionMode(self.lstSub.ExtendedSelection)
        self.lstProd = QtGui.QListWidget()
        self.lstProd.setSelectionMode(self.lstProd.ExtendedSelection)
        # populate lists
        cntr = 0
        for m_ in self._mlist:
            cntr += 1
            name = 'None'
            comp = 'None'
            if m_ in self._ccdict:
                comp = self._ccdict[m_]
            if m_ in self._cndict:
                name = self._cndict[m_]
            item = QtGui.QListWidgetItem(m_)
            item.setToolTip('{}\t{}'.format(name, comp))
            if m_ in self._flist:
                item.setForeground(self._fixColour)
            self.lstSub.addItem(item.clone())
            self.lstProd.addItem(item)
            # if cntr == 20: break

        # create buttons
        self.btAddSub = QtGui.QPushButton('Add substrate(s)')
        QtCore.QObject.connect(
            self.btAddSub, QtCore.SIGNAL('clicked()'), self.addSelectedSubstrates
        )
        self.btAddProd = QtGui.QPushButton('Add product(s)')
        QtCore.QObject.connect(
            self.btAddProd, QtCore.SIGNAL('clicked()'), self.addSelectedProducts
        )
        self.btReverse = QtGui.QPushButton('Reversible')
        self.btReverse.setCheckable(True)
        self.btReverse.setChecked(True)

        # create tables
        self.tblSub = QtGui.QTableWidget()
        self.tblSub.setSortingEnabled(True)
        self.tblSub.insertColumn(0)
        self.tblSub.insertColumn(1)
        self.tblSub.setHorizontalHeaderLabels(
            QtCore.QStringList(('Coefficient', 'Metabolite'))
        )
        self.tblSub.verticalHeader().setVisible(False)
        QtCore.QObject.connect(
            self.tblSub, QtCore.SIGNAL('cellChanged(int,int)'), self.checkBalance
        )
        self.tblSubRow = 0
        self.tblProd = QtGui.QTableWidget()
        self.tblProd.setSortingEnabled(True)
        self.tblProd.insertColumn(0)
        self.tblProd.insertColumn(1)
        self.tblProd.setHorizontalHeaderLabels(
            QtCore.QStringList(('Coefficient', 'Metabolite'))
        )
        self.tblProd.verticalHeader().setVisible(False)
        self.tblProdRow = 0
        QtCore.QObject.connect(
            self.tblProd, QtCore.SIGNAL('cellChanged(int,int)'), self.checkBalance
        )

        self.tblBal = QtGui.QTableWidget()
        self.tblBal.setMaximumHeight(150)
        self.tblBal.insertRow(0)
        self.tblBal.insertRow(1)
        self.tblBal.insertRow(2)
        self.tblBal.verticalHeader().setVisible(False)
        self.tblBalCol = 0

        # set up menu and status bar
        menuBar = QtGui.QMenuBar()

        exitAction = QtGui.QAction('&Exit', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.setStatusTip('Exit and loose changes')
        exitAction.triggered.connect(QtGui.qApp.quit)
        addReAction = QtGui.QAction('&Add reaction and exit', self)
        addReAction.setShortcut('Ctrl+A')
        addReAction.setStatusTip('Add reaction to model and exit')
        addReAction.triggered.connect(self.getNewReaction)

        fileMenu = menuBar.addMenu('&Model')
        fileMenu.addAction(exitAction)
        fileMenu.addAction(addReAction)

        subAddAction = QtGui.QAction('&Add', self)
        subAddAction.triggered.connect(self.addSelectedSubstrates)
        subDelAction = QtGui.QAction('&Delete selected', self)
        subDelAction.triggered.connect(self.deleteSubstrates)
        subDelAllAction = QtGui.QAction('&Delete all', self)
        subDelAllAction.triggered.connect(self.deleteAllSubstrates)
        subMenu = menuBar.addMenu('&Substrate')
        subMenu.addAction(subAddAction)
        subMenu.addAction(subDelAction)
        subMenu.addAction(subDelAllAction)

        prodAddAction = QtGui.QAction('&Add', self)
        prodAddAction.triggered.connect(self.addSelectedProducts)
        prodDelAction = QtGui.QAction('&Delete selected', self)
        prodDelAction.triggered.connect(self.deleteProducts)
        prodDelAllAction = QtGui.QAction('&Delete all', self)
        prodDelAllAction.triggered.connect(self.deleteAllProducts)
        prodMenu = menuBar.addMenu('&Product')
        prodMenu.addAction(prodAddAction)
        prodMenu.addAction(prodDelAction)
        prodMenu.addAction(prodDelAllAction)

        self.statusBar = QtGui.QStatusBar()
        self.statusBar.showMessage('{} ready'.format('Model'))

        # do layout
        grid = QtGui.QGridLayout()
        grid.setSpacing(10)

        grid.addWidget(menuBar, 0, 0, 1, 2)
        grid.addWidget(self.statusBar, 0, 2, 1, 2)
        grid.addWidget(lblId, 1, 0)
        grid.addWidget(self.txtId, 1, 1)
        grid.addWidget(lblNm, 1, 2)
        grid.addWidget(self.txtNm, 1, 3)
        grid.addWidget(lblSub, 2, 0, 1, 2)
        grid.addWidget(lblProd, 2, 2, 1, 2)
        grid.addWidget(self.lstSub, 3, 0, 1, 2)
        grid.addWidget(self.lstProd, 3, 2, 1, 2)
        grid.addWidget(self.btAddSub, 4, 0)
        grid.addWidget(self.btReverse, 4, 1, 1, 2)
        grid.addWidget(self.btAddProd, 4, 3)
        grid.addWidget(self.tblSub, 5, 0, 1, 2)
        grid.addWidget(self.tblProd, 5, 2, 1, 2)
        grid.addWidget(self.tblBal, 6, 0, 1, 4)

        self.setLayout(grid)
        self.setGeometry(self.mousePos.x() - 75, self.mousePos.y() - 75, 500, 640)
        self.setWindowTitle('Reaction Creator')
        self.show()


def createReaction(mod):
    """
    Create a reaction using the graphical Reaction Creator

     - *mod* a CBMPy model object

    """
    cfdict = {}
    cndict = {}
    ccdict = {}
    for s_ in mod.species:
        cfdict[s_.getId()] = s_.getChemFormula()
        cndict[s_.getId()] = s_.getName()
        ccdict[s_.getId()] = s_.compartment
    app = QtGui.QApplication([])
    ex = ReactionCreator(
        mod.getReactionIds(),
        mod.getSpeciesIds(),
        mod.getBoundarySpeciesIds(),
        cfdict,
        cndict,
        ccdict,
    )
    app.exec_()
    newR = ex.NewReaction
    del app, ex
    if newR == None:
        return None
    else:
        R = CBMReaction(newR['id'], name=newR['name'], reversible=newR['reversible'])
        R.is_balanced = newR['is_balanced']
        R.is_exchange = newR['is_exchange']
        for r_ in newR['reagents']:
            R.createReagent(r_, newR['reagents'][r_])
            mod.getSpecies(r_).setReagentOf(newR['id'])
        mod.addReaction(R, create_default_bounds=False)
        mod.createReactionLowerBound(newR['id'], newR['lower_bound'])
        mod.createReactionUpperBound(newR['id'], newR['upper_bound'])
        return R


class CBFileDialogue(QtGui.QWidget):
    _appTitle = 'Open file'
    work_dir = None
    model_file = None
    mode = None

    def __init__(self, work_dir, mode='open', filters=None):
        super(CBFileDialogue, self).__init__()
        self.mousePos = self.cursor().pos()
        self.work_dir = work_dir
        self.mode = mode
        if mode == 'save':
            self._appTitle = 'Save file'
        self.initUI()

    def initUI(self):
        # self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        # self.setHidden(True)
        self.__dlg__ = QtGui.QFileDialog(self)
        if self.mode == 'open':
            # self.model_file = str(self.__dlg__.getOpenFileName(self, 'Open file', self.work_dir, options=QtGui.QFileDialog.DontUseNativeDialog))
            self.model_file = str(
                self.__dlg__.getOpenFileName(self, 'Open file', self.work_dir)
            )
        elif self.mode == 'save':
            # self.model_file = str(self.__dlg__.getSaveFileName(self, 'Save file as', self.work_dir, options=QtGui.QFileDialog.DontUseNativeDialog))
            self.model_file = str(
                self.__dlg__.getSaveFileName(self, 'Save file as', self.work_dir)
            )
        self.model_file = os.path.normpath(self.model_file)


def fileDialogue(work_dir=None, mode='open', filters=None):
    if work_dir == None:
        work_dir = os.getcwd()
    if mode in ['open', 'save']:
        app = QtGui.QApplication([])
        fileApp = CBFileDialogue(work_dir, mode=mode, filters=filters)
        model_file = fileApp.model_file
        fileApp.__dlg__.done(1)
        app.exit()
        if mode == 'open':
            return model_file
        else:
            return True


class ViewSVG(QtGui.QWidget):
    _fixColour = None
    _errColour = None
    _goodColour = None
    _appTitle = 'ViewSVG'

    def __init__(self, filename):
        super(ViewSVG, self).__init__()
        self.mousePos = self.cursor().pos()
        self._fixColour = QtGui.QColor(0, 0, 153, alpha=255)
        self._errColour = QtGui.QColor(255, 0, 0, alpha=255)
        self._goodColour = QtGui.QColor(0, 100, 0, alpha=255)
        self.filename = os.path.abspath(filename)
        print('\nViewing file: {}'.format(filename))
        self.initUI()

    def initUI(self):
        # create panels
        self.txtId = QtGui.QTextEdit()

        # set up menu and status bar
        menuBar = QtGui.QMenuBar()

        exitAction = QtGui.QAction('&Exit', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.setStatusTip('Exit and loose changes')
        exitAction.triggered.connect(QtGui.qApp.quit)

        fileMenu = menuBar.addMenu('&File')
        fileMenu.addAction(exitAction)

        self.statusBar = QtGui.QStatusBar()
        self.statusBar.showMessage('Ready ...')

        # webkit panel
        scene = QtGui.QGraphicsScene()
        view = QtGui.QGraphicsView(scene)

        br = QtSvg.QGraphicsSvgItem(self.filename).boundingRect()

        webview = QGraphicsWebView()
        # webview.load(QtCore.QUrl("C:\your_interactive_svg.svg"))
        webview.load(QtCore.QUrl(QtCore.QUrl.fromLocalFile(self.filename)))
        webview.setFlags(QtGui.QGraphicsItem.ItemClipsToShape)
        webview.setCacheMode(QtGui.QGraphicsItem.NoCache)
        webview.resize(br.width(), br.height())

        scene.addItem(webview)
        view.resize(br.width() + 10, br.height() + 10)
        # view.show()

        # do layout
        grid = QtGui.QGridLayout()
        grid.setSpacing(10)

        grid.addWidget(menuBar, 0, 0, 1, 2)
        # grid.addWidget(self.txtId, 1, 0, 1, 2)
        grid.addWidget(view, 1, 0, 4, 4)
        grid.addWidget(self.statusBar, 5, 0, 1, 4)

        self.setLayout(grid)
        self.setGeometry(self.mousePos.x() - 75, self.mousePos.y() - 75, 500, 640)
        self.setWindowTitle(self._appTitle)
        self.show()


def loadViewSVG(filename):
    app = QtGui.QApplication([])
    ex = ViewSVG(filename)
    app.exec_()


class ValueSlider(QtGui.QWidget):
    _fixColour = None
    _errColour = None
    _goodColour = None
    _appTitle = 'ValueSlider'

    def __init__(self):
        super(ValueSlider, self).__init__()
        self.mousePos = self.cursor().pos()
        self._fixColour = QtGui.QColor(0, 0, 153, alpha=255)
        self._errColour = QtGui.QColor(255, 0, 0, alpha=255)
        self._goodColour = QtGui.QColor(0, 100, 0, alpha=255)
        self.initUI()

    def initUI(self):

        # create panels
        # self.txtId = QtGui.QTextEdit()

        l1a = QtGui.QLabel(self)
        l1a.setText('Property')
        sld1 = QtGui.QSlider(QtCore.Qt.Horizontal, self)
        sld1.setTickPosition(sld1.TicksBelow)

        sld1_min = -100
        sld1_max = 100

        sld1.setMinimum(sld1_min)
        sld1.setMaximum(sld1_max)
        sld1.setTickInterval((sld1_min - sld1_max) / 10.0)
        sld1.setSingleStep(0.1)

        sld1.setFocusPolicy(QtCore.Qt.NoFocus)
        sld1.valueChanged[int].connect(self.changeValue)
        self.l1b = QtGui.QLabel(self)
        self.l1b.setText('0.0')

        # do layout
        grid = QtGui.QGridLayout()
        grid.setSpacing(10)

        grid.addWidget(l1a, 0, 0, 1, 1)
        grid.addWidget(self.l1b, 0, 1, 1, 1)
        grid.addWidget(sld1, 0, 2, 1, 5)

        # grid.addWidget(self.txtId, 1, 0, 1, 2)
        # grid.addWidget(menuBar, 0, 0)
        # grid.addWidget(self.statusBar, 0, 1)

        self.setLayout(grid)
        self.setGeometry(self.mousePos.x() - 75, self.mousePos.y() - 75, 280, 170)
        self.setWindowTitle(self._appTitle)
        self.show()

    def changeValue(self, value):
        getattr(self, 'l1b').setText('{}'.format(value))


def loadSlider():
    app = QtGui.QApplication([])
    ex = ValueSlider()
    app.exec_()


data = "<DATASTART><return>{}</return>"

if __name__ == '__main__':
    print(os.sys.argv)
    if os.sys.argv[1] == 'fileOpen':
        filename = fileDialogue(work_dir=None, mode='open', filters=None)
        print(data.format(filename))
        os.sys.exit(0)

    # subprocess.check_output(['python', '_qtloader.py', 'fileOpen']).split('<DATASTART>')[1].strip()


# template widget
"""
    class SmallAppBasicGrid(QtGui.QWidget):
        _fixColour = None
        _errColour = None
        _goodColour = None
        _appTitle = 'SmallAppBaseGrid'

        def __init__(self):
            super(SmallAppBasicGrid, self).__init__()
            self.mousePos = self.cursor().pos()
            self._fixColour = QtGui.QColor(0,0,153,alpha=255)
            self._errColour = QtGui.QColor(255,0,0,alpha=255)
            self._goodColour = QtGui.QColor(0,100,0,alpha=255)
            self.initUI()


        def initUI(self):

            # create panels
            self.txtId = QtGui.QTextEdit()

            # set up menu and status bar
            menuBar = QtGui.QMenuBar()

            exitAction = QtGui.QAction('&Exit', self)
            exitAction.setShortcut('Ctrl+Q')
            exitAction.setStatusTip('Exit and loose changes')
            exitAction.triggered.connect(QtGui.qApp.quit)

            fileMenu = menuBar.addMenu('&File')
            fileMenu.addAction(exitAction)

            self.statusBar = QtGui.QStatusBar()
            self.statusBar.showMessage('Ready ...')

            # do layout
            grid = QtGui.QGridLayout()
            grid.setSpacing(10)

            grid.addWidget(menuBar, 0, 0, 1, 2)
            grid.addWidget(self.txtId, 1, 0, 1, 2)
            grid.addWidget(self.statusBar, 2, 0, 1, 2)

            self.setLayout(grid)
            self.setGeometry(self.mousePos.x()-75, self.mousePos.y()-75, 500, 640)
            self.setWindowTitle(self._appTitle)
            self.show()

    def loadBasicApp(mod):
        app = QtGui.QApplication([])
        ex = SmallAppBasicGrid()
        app.exec_()
"""

# template microGUI
"""
    class MicroGUI(QtGui.QWidget):
        _appTitle = 'MicroGUI'

        def __init__(self):
            super(MicroGUI, self).__init__()
            self.mousePos = self.cursor().pos()
            self.work_dir = work_dir
            self.initUI()


        def initUI(self):
            self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
            self.setHidden(True)

            # action code (this example is for opening a file dialogue)
            self.model_file = str(QtGui.QFileDialog.getOpenFileName(self, 'Open file', os.getcwd()))



    def loadMicroGUI(*args):
        app = QtGui.QApplication([])
        mGUI = OpenFileDialogue()
        appTitle = mGUI._appTitle
        del mGUI, app
        return
"""

"""
import sys
from PyQt4 import QtCore, QtGui, QtSvg
from PyQt4.QtWebKit import QGraphicsWebView
if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)

    scene = QtGui.QGraphicsScene()
    view = QtGui.QGraphicsView(scene)

    br = QtSvg.QGraphicsSvgItem("C:\your_interactive_svg.svg").boundingRect()

    webview = QGraphicsWebView()
    webview.load(QtCore.QUrl("C:\your_interactive_svg.svg"))
    webview.load(QtCore.QUrl(QtCore.QUrl.fromLocalFile("C:\your_interactive_svg.svg")))
    webview.setFlags(QtGui.QGraphicsItem.ItemClipsToShape)
    webview.setCacheMode(QtGui.QGraphicsItem.NoCache)
    webview.resize(br.width(), br.height())

    scene.addItem(webview)
    view.resize(br.width()+10, br.height()+10)
    view.show()
    sys.exit(app.exec_())
"""



"""
CBMPy: CBGUI module
===================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBGUI.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import os

__HAVE_QT4__ = False
__HAVE_WX__ = False
try:
    from . import CBWx

    __HAVE_WX__ = True
except ImportError as ex:
    print('WARNING: Error importing wxpython module')
    # print(ex)
# try:
# if os.sys.platform == 'win32':
# from . import CBQt4
# __HAVE_QT4__ = True
# else:
# print('\nINFO: Qt currently only available on Windows')
# except ImportError as ex:
# print('WARNING: Error importing CBQt4 module')
# print(ex)

if __HAVE_WX__:

    def loadCBGUI(mod, version=2):
        """
        Load an FBA model instance into the quick editor to view or change basic model properties

         - *mod* a PySCeS CBMPy model instance

        """
        if version == 2:
            loadCBGUI = CBWx.runModelEditor(mod)
        elif version == 3:
            CBWx.runMyAUIApp()


else:
    print(
        '\nWX GUI is not available please make sure WxPython is installed (http://www.wxpython.org)'
    )

if __HAVE_QT4__:

    def createReaction(mod):
        """
        Load the QT4 reaction creator widget

         - *mod* a PySCeS CBMPy model instance

        """
        CBQt4.createReaction(mod)

    def openFileName(work_dir=None):
        """
        Load the QT4 file open selection dialogue

         - *work_dir* the optional initial directory

        """
        if work_dir == None:
            work_dir = os.getcwd()
        return CBQt4.fileDialogue(work_dir, 'open', None)

    def saveFileName(work_dir=None):
        """
        Load the QT4 file save selection dialogue

         - *work_dir* the optional initial directory

        """
        if work_dir == None:
            work_dir = os.getcwd()
        return CBQt4.fileDialogue(work_dir, 'save', None)


else:
    print('\nQT4 GUI tools are not available please make sure PyQT4 is installed ()')


'''
def loadBasicApp(mod=None):
    """
    Load the QT4 basic app widget

     - *mod* a PySCeS CBMPy model instance

    """
    if __HAVE_QT4__:
        CBQt4.loadBasicApp(mod)
    else:
        print('\nQT4 GUI tools are not available please make sure PyQT4 is installed ()')
'''



"""
CBMPy: CBGLPK module
====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBGLPK.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

from packaging import version as pkgver

HAVE_SYMPY = False
try:
    import sympy

    if pkgver.parse(sympy.__version__) >= pkgver.Version('0.7.5'):
        HAVE_SYMPY = True

    else:
        del sympy
        print(
            '\nWARNING: SymPy version 0.7.5 or newer is required for symbolic matrix support.'
        )
except ImportError:
    HAVE_SYMPY = False


import os, time
import numpy
from . import CBWrite
from .CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']


HAVE_GLPK = False
GLPK_SOLUTION_STATUS = None

try:
    import glpk

    lp = glpk.LPX()
    HAVE_GLPK = True
except:
    raise ImportError

# configuration options for GLPK
GLPK_CFG = {
    'simplex': {
        'meth': glpk.LPX.PRIMAL,
        #'meth' : glpk.LPX.DUAL,
        #'meth' : glpk.LPX.DUALP,
        'tol_bnd': 1.0e-6,
        'tol_dj': 1.0e-6,
        'tol_piv': 1.0e-10,
    }
}

GLPK_STATUS = {
    1: 'LPS_UNDEF',
    2: 'LPS_FEAS',
    3: 'LPS_INFEAS',
    4: 'LPS_NOFEAS',
    5: 'LPS_OPT',
    6: 'LPS_UNBND',
}

GLPK_STATUS2 = {
    'opt': 'LPS_OPT',
    'undef': 'LPS_UNDEF',
    'feas': 'LPS_FEAS',
    'infeas': 'LPS_INFEAS',
    'nofeas': 'LPS_NOFEAS',
    'unbnd': 'LPS_UNBND',
}


GLPK_SILENT_MODE = True
GLPK_INFINITY = 1.0e9


def glpk_constructLPfromFBA(fba, fname=None):
    """
    Create a GLPK LP in memory.
    - *fba* an FBA object
    - *fname* optional filename if defined writes out the constructed lp

    """
    _Stime = time.time()

    # define model and add variables
    lp = glpk.LPX()
    lp.name = fba.getId()
    lp.cols.add(len(fba.N.col))

    if HAVE_SYMPY and fba.N.__array_type__ == sympy.MutableDenseMatrix:
        print('INFO: GLPK requires floating point, converting N')
        Nmat = numpy.array(fba.N.array).astype('float')
        RHSmat = numpy.array(fba.N.RHS).astype('float')
        if fba.CM != None:
            CMmat = numpy.array(fba.CM.array).astype('float')
            CMrhs = numpy.array(fba.CM.RHS).astype('float')
    else:
        Nmat = fba.N.array
        RHSmat = fba.N.RHS
        if fba.CM != None:
            CMmat = fba.CM.array
            CMrhs = fba.CM.RHS

    varMap = {}
    for n_ in range(Nmat.shape[1]):
        varMap[fba.N.col[n_]] = n_
        lp.cols[n_].name = fba.N.col[n_]

    # print varMap

    try:
        # define objective
        osense = fba.getActiveObjective().operation.lower()
        if osense in ['minimize', 'minimise', 'min']:
            lp.obj.maximize = False
        elif osense in ['maximize', 'maximise', 'max']:
            lp.obj.maximize = True
        else:
            raise RuntimeError('\n%s - is not a valid objective operation' % osense)
        lp.obj.name = fba.getActiveObjective().getId()
        for fo_ in fba.getActiveObjective().fluxObjectives:
            lp.obj[varMap[fo_.reaction]] = fo_.coefficient
    except AttributeError:
        print('\nWARNING(GLPK create LP): no objective function defined')
    # create N constraints
    lp.rows.add(Nmat.shape[0])
    # conMap = {}
    # for n_ in range(Nmat.shape[0]):
    ##conMap[fba.N.row[n_]] = n_
    # lp.rows[n_].name = fba.N.row[n_]

    # tnew = time.time()
    for r_ in range(Nmat.shape[0]):
        # name and coefficients
        newCon = []
        for c_ in range(Nmat.shape[1]):
            newCon.append((c_, Nmat[r_, c_]))
        lp.rows[r_].name = fba.N.row[r_]
        lp.rows[r_].matrix = newCon

        # sense and rhs
        rhs = RHSmat[r_]
        if fba.N.operators[r_] in ['<=', '<', 'L']:
            lp.rows[r_].bounds = None, rhs
        elif fba.N.operators[r_] in ['>=', '>', 'G']:
            lp.rows[r_].bounds = rhs, None
        elif fba.N.operators[r_] in ['=', 'E']:
            lp.rows[r_].bounds = rhs
        else:
            raise RuntimeError('INFO: invalid operator: %s' % fba.N.operators[n])

    # add user defined constraints
    if fba.CM != None and fba.CM.shape[0] > 0:
        baseRows = len(lp.rows)
        lp.rows.add(CMmat.shape[0])
        for r_ in range(CMmat.shape[0]):
            # name and coefficients
            newCon = []
            for c_ in range(CMmat.shape[1]):
                newCon.append((c_, CMmat[r_, c_]))
            lp.rows[baseRows + r_].name = fba.CM.row[r_]
            lp.rows[baseRows + r_].matrix = newCon

            # sense and rhs
            rhs = CMrhs[r_]
            if fba.CM.operators[r_] in ['<=', '<', 'L']:
                lp.rows[baseRows + r_].bounds = None, rhs
            elif fba.CM.operators[r_] in ['>=', '>', 'G']:
                lp.rows[baseRows + r_].bounds = rhs, None
            elif fba.CM.operators[r_] in ['=', 'E']:
                lp.rows[baseRows + r_].bounds = rhs
            else:
                raise RuntimeError('INFO: invalid operator: %s' % fba.N.operators[n])

    # add bounds
    for r_ in fba.reactions:
        lb = ub = None
        lb = fba.getReactionLowerBound(r_.getId())
        ub = fba.getReactionUpperBound(r_.getId())

        if lb in ['Infinity', 'inf', 'Inf', 'infinity']:
            lb = GLPK_INFINITY
        elif lb in ['-Infinity', '-inf', '-Inf', '-infinity', None]:
            lb = -GLPK_INFINITY
        elif numpy.isinf(lb):
            if lb < 0.0:
                lb = -GLPK_INFINITY
            else:
                lb = GLPK_INFINITY
        if ub in ['Infinity', 'inf', 'Inf', 'infinity', None]:
            ub = GLPK_INFINITY
        elif ub in ['-Infinity', '-inf', '-Inf', '-infinity']:
            ub = -GLPK_INFINITY
        elif numpy.isinf(ub):
            if ub < 0.0:
                ub = -GLPK_INFINITY
            else:
                ub = GLPK_INFINITY

        if ub != GLPK_INFINITY and lb != -GLPK_INFINITY and ub == lb:
            lp.cols[varMap[r_.getId()]].bounds = lb
        elif ub != GLPK_INFINITY and lb != -GLPK_INFINITY:
            lp.cols[varMap[r_.getId()]].bounds = lb, ub
        elif ub != GLPK_INFINITY:
            lp.cols[varMap[r_.getId()]].bounds = None, ub
        elif lb != -GLPK_INFINITY:
            lp.cols[varMap[r_.getId()]].bounds = lb, None
        else:
            lp.cols[varMap[r_.getId()]].bounds = None

    print('\ngplk_constructLPfromFBA time: {}\n'.format(time.time() - _Stime))
    if fname != None:
        lp.write(cpxlp=fname + '.lp')
    return lp


def glpk_Solve(lp, method='s'):
    """
    Solve the LP and create a status attribute with the solution status

     - *method* [default='s'] 's' = simplex, 'i' = interior, 'e' = exact

    GLPK solver options can be set in the dictionary GLPK_CFG

    """

    if method == 'i':
        glpksol = lp.interior()
    elif method == 'e':
        glpksol = lp.exact()
    else:
        glpksol = lp.simplex(**GLPK_CFG['simplex'])

    global GLPK_SOLUTION_STATUS
    GLPK_SOLUTION_STATUS = glpk_getSolutionStatus(lp)

    # if status == 'LPS_UNDEF':
    # sd.presolve = False
    # status = glpk.glp_simplex(lp, sd)
    # if status == 'LPS_UNDEF':
    # print('\nINFO: Primal solver failure switching to dual-primal solver\n')
    # sd.presolve = False
    # sd.obj_ul = 1.0e10
    # sd.obj_ll = 1.0e10
    # sd.meth = glpk.GLP_DUALP
    # glpk.glp_simplex(lp, sd)
    # status = glpk_getSolutionStatus(lp)

    if GLPK_SOLUTION_STATUS in [
        'LPS_UNDEF',
        'LPS_FEAS',
        'LPS_INFEAS',
        'LPS_NOFEAS',
        'LPS_OPT',
        'LPS_UNBND',
    ]:
        if not GLPK_SILENT_MODE:
            print('Solution status returned as: {}'.format(GLPK_SOLUTION_STATUS))
            print("Objective value = ", lp.obj.value)
        return GLPK_SOLUTION_STATUS
    else:
        print("INFO: No solution available ({})".format(GLPK_SOLUTION_STATUS))
        return None


def glpk_getSolutionStatus(lp):
    """
    Returns one of:

     - *LPS_OPT*: solution is optimal;
     - *LPS_FEAS*: solution is feasible;
     - *LPS_INFEAS*: solution is infeasible;
     - *LPS_NOFEAS*: problem has no feasible solution;
     - *LPS_UNBND*: problem has unbounded solution;
     - *LPS_UNDEF*: solution is undefined.

    """
    return GLPK_STATUS2[lp.status]


def glpk_analyzeModel(
    f,
    lpFname=None,
    return_lp_obj=False,
    with_reduced_costs='unscaled',
    with_sensitivity=False,
    del_intermediate=False,
    build_n=True,
    quiet=False,
    oldlpgen=False,
    method='s',
):
    """
    Optimize a model and add the result of the optimization to the model object
    (e.g. `reaction.value`, `objectiveFunction.value`). The stoichiometric
    matrix is automatically generated. This is a common function available in all
    solver interfaces. By default returns the objective function value

     - *f* an instantiated PySCeSCBM model object
     - *lpFname* [default=None] the name of the intermediate LP file saved when this has a string value.
     - *return_lp_obj* [default=False] off by default when enabled it returns the PyGLPK LP object
     - *with_reduced_costs* [default='unscaled'] calculate and add reduced cost information to mode this can be: 'unscaled' or 'scaled'
       or anything else which is interpreted as 'None'. Scaled means s_rcost = (r.reduced_cost*rval)/obj_value
     - *with_sensitivity* [default=False] add solution sensitivity information (not yet implemented)
     - *del_intermediate* [default=False] delete the intermediary files after updating model object, useful for server applications
     - *build_n* [default=True] generate stoichiometry from the reaction network (reactions/reagents/species)
     - *quiet* [default=False] suppress glpk output
     - *method* [default='s'] select the GLPK solver method, see the GLPK documentation for details

       - 's': simplex
       - 'i': interior
       - 'e': exact

    """

    if build_n:
        f.buildStoichMatrix()

    if f.__check_gene_activity__:
        f.updateNetwork(lower=0.0, upper=0.0)

    fid = f.id

    if with_reduced_costs == 'scaled':
        f.SCALED_REDUCED_COSTS = True
    elif with_reduced_costs == 'unscaled':
        f.SCALED_REDUCED_COSTS = False
    else:
        f.SCALED_REDUCED_COSTS = None

    if lpFname == None:
        flp = glpk_constructLPfromFBA(f, fname=None)
        f.id = '_glpktmp_.tmp'
    else:
        flp = glpk_constructLPfromFBA(f, fname=lpFname)
        fid = lpFname

    _Stime = time.time()

    print('\nglpk_analyzeModel FBA --> LP time: {}\n'.format(time.time() - _Stime))
    f.id = fid
    glpk_Solve(flp, method)

    glpk_setFBAsolutionToModel(f, flp, with_reduced_costs=with_reduced_costs)
    glpk_setSolutionStatusToModel(f, flp)

    if oldlpgen and del_intermediate:
        os.remove(LPF)
    objv = f.getActiveObjective().getValue()
    print('\nanalyzeModel objective value: {}\n'.format(objv))
    if return_lp_obj:
        return flp
    else:
        del flp
        return objv


def glpk_setSolutionStatusToModel(m, lp):
    """
    Sets the lp solutions status to the CBMPy model

    """
    m.SOLUTION_STATUS = glpk_getSolutionStatus(lp)
    # TODO: need to synchronise all solvers to same integer system
    if m.SOLUTION_STATUS == 'LPS_OPT':
        m.SOLUTION_STATUS_INT = 1
    else:
        m.SOLUTION_STATUS_INT = 999


def glpk_setFBAsolutionToModel(fba, lp, with_reduced_costs='unscaled'):
    """
    Sets the FBA solution from a CPLEX solution to an FBA object

     - *fba* and fba object
     - *lp* a CPLEX LP object
     - *with_reduced_costs* [default='unscaled'] calculate and add reduced cost information to mode this can be: 'unscaled' or 'scaled'
       or anything else which is interpreted as None. Scaled is: s_rcost = (r.reduced_cost*rval)/obj_value

    """
    sol, objname, objval = glpk_getOptimalSolution(lp)
    if glpk_getSolutionStatus(lp) == 'LPS_OPT':
        (
            fba.objectives[fba.activeObjIdx].solution,
            fba.objectives[fba.activeObjIdx].value,
        ) = (sol, objval)
    else:
        (
            fba.objectives[fba.activeObjIdx].solution,
            fba.objectives[fba.activeObjIdx].value,
        ) = (sol, numpy.NaN)
    for r in fba.reactions:
        rid = r.getId()
        if rid in sol:
            r.value = sol[rid]
        else:
            r.value = None
    scaled = False
    if with_reduced_costs == 'scaled':
        scaled = True
        with_reduced_costs = True
    elif with_reduced_costs == 'unscaled':
        scaled = False
        with_reduced_costs = True
    else:
        with_reduced_costs = False

    if objval != None and sol != {} and with_reduced_costs:
        RC = glpk_getReducedCosts(lp, scaled=scaled)
        setReducedCosts(fba, RC)
    else:
        setReducedCosts(fba, {})


def glpk_getOptimalSolution(c):
    """
    From a GLPK model extract a tuple of solution, ObjFuncName and ObjFuncVal
    """
    s_val = []
    s_name = []
    fba_sol = {}
    objf_name = None
    objf_val = None
    try:
        objf_name = c.obj.name
        objf_val = c.obj.value
        for n in c.cols:
            fba_sol.update({n.name: n.value})
    except Exception as ex:
        print(ex)
        print('WARNING: No solution to get')
        s_val = []
        s_name = []
        fba_sol = {}
        objf_val = None
    del s_name, s_val
    return fba_sol, objf_name, objf_val


def glpk_getReducedCosts(c, scaled=False):
    """
    Extract ReducedCosts from LP and return as a dictionary 'Rid' : reduced cost

     - *c* a GLPK LP object
     - *scaled* scale the reduced cost by the optimal flux value

    """
    s_name = []
    r_costs = []
    s_val = []

    for c_ in c.cols:
        s_name.append(c_.name)
        r_costs.append(c_.dual)
        s_val.append(c_.value)
    objf_val = c.obj.value
    output = {}
    for v in range(len(s_name)):
        if scaled:
            try:
                r_val = r_costs[v] * s_val[v] / objf_val
            except:
                r_val = float('nan')
        else:
            r_val = r_costs[v]
        output.update({s_name[v]: r_val})
    del s_name, r_costs, s_val
    return output


def glpk_FluxVariabilityAnalysis(
    fba,
    selected_reactions=None,
    pre_opt=True,
    tol=None,
    objF2constr=True,
    rhs_sense='lower',
    optPercentage=100.0,
    work_dir=None,
    quiet=True,
    debug=False,
    oldlpgen=False,
    markupmodel=True,
    default_on_fail=False,
    roundoff_span=10,
    method='s',
):
    """
    Perform a flux variability analysis on an fba model:

     - *fba* an FBA model object
     - *selected reactions* [default=None] means use all reactions otherwise use the reactions listed here
     - *pre_opt* [default=True] attempt to presolve the FBA and report its results in the ouput, if this is disabled and *objF2constr* is True then the vid/value of the current active objective is used
     - *tol*  [default=None] do not floor/ceiling the objective function constraint, otherwise round of to *tol*
     - *rhs_sense* [default='lower'] means objC >= objVal the inequality to use for the objective constraint can also be *upper* or *equal*
     - *optPercentage* [default=100.0] means the percentage optimal value to use for the RHS of the objective constraint: optimal_value*(optPercentage/100.0)
     - *work_dir* [default=None] the FVA working directory for temporary files default = cwd+fva
     - *debug* [default=False] if True write out all the intermediate FVA LP's into work_dir
     - *quiet* [default=False] if enabled supress CPLEX output
     - *objF2constr* [default=True] add the model objective function as a constraint using rhs_sense etc. If
       this is True with pre_opt=False then the id/value of the active objective is used to form the constraint
     - *markupmodel* [default=True] add the values returned by the fva to the reaction.fva_min and reaction.fva_max
     - *default_on_fail* [default=False] if *pre_opt* is enabled replace a failed minimum/maximum with the solution value
     - *roundoff_span* [default=10] number of digits is round off (not individual min/max values)
     - *method* [default='s'] select the GLPK solver method, see the GLPK documentation for details

       - 's': simplex
       - 'i': interior
       - 'e': exact


    Returns an array with columns Reaction, Reduced Costs, Variability Min, Variability Max, abs(Max-Min), MinStatus, MaxStatus and a list containing the row names.

    """
    if work_dir == None:
        work_dir = os.getcwd()
    else:
        assert os.path.exists(work_dir), '\nWhat did you think would happen now!'
    if debug:
        debug_dir = os.path.join(work_dir, 'DEBUG')
        if not os.path.exists(debug_dir):
            os.mkdir(debug_dir)
    s2time = time.time()
    # generate a presolution
    print('GLPK is using solver option: "{}"'.format(method))
    cpx = OPTIMAL_PRESOLUTION = None
    pre_sol = pre_oid = pre_oval = None
    REDUCED_COSTS = {}
    (
        cpx,
        pre_sol,
        pre_oid,
        pre_oval,
        OPTIMAL_PRESOLUTION,
        REDUCED_COSTS,
    ) = glpk_func_GetCPXandPresolve(
        fba, pre_opt, objF2constr, quiet=quiet, oldlpgen=oldlpgen, method=method
    )

    # if required add the objective function as a constraint
    if objF2constr:
        glpk_func_SetObjectiveFunctionAsConstraint(
            cpx, rhs_sense, pre_oval, tol, optPercentage
        )
    if debug:
        cpx.write(cpxlp=os.path.join(debug_dir, 'FVA_base.lp'))

    # do the FVA
    NUM_FLX = len(fba.reactions)
    print('Total number of reactions: {}'.format(NUM_FLX))
    if selected_reactions != None:
        rids = fba.getReactionIds()
        for r in selected_reactions:
            assert r in rids, "\n%s is not a valid reaction name" % r
    else:
        selected_reactions = fba.getReactionIds()
    NUM_FLX = len(selected_reactions)
    print('Number of user selected variables: {}'.format(NUM_FLX))
    try:
        OUTPUT_ARRAY = numpy.zeros((NUM_FLX, 7), numpy.double)
    except AttributeError:
        OUTPUT_ARRAY = numpy.zeros((NUM_FLX, 7))
    OUTPUT_NAMES = []
    cntr = 0
    tcnt = 0
    # this is a memory hack --> prevents solver going berserk
    mcntr = 0
    mcntr_cntrl = 3
    mps_filename = '_{}_.mps'.format(str(time.time()).replace('.', '_'))
    cpx.write(mps=mps_filename)
    for Ridx in range(NUM_FLX):
        R = selected_reactions[Ridx]
        OUTPUT_NAMES.append(R)
        max_error_iter = 1
        GOMIN = True
        gomin_cntr = 0
        while GOMIN:
            MIN_STAT = 0
            # MIN
            # TODO: bgoli: see whether this also works with 'minimize'
            glpk_setObjective(cpx, 'min%s' % R, [(1, R)], 'min', reset=True)
            ##  cplx_setBounds(c, id, min=None, max=None) # think about this
            MIN_STAT = glpk_Solve(cpx, method=method)
            if MIN_STAT == 'LPS_OPT':
                MIN_STAT = 1
            elif MIN_STAT == 'LPS_UNBND':
                MIN_STAT = 2
            elif MIN_STAT == 'LPS_NOFEAS':
                MIN_STAT = 3
            else:
                MIN_STAT = 4
            if MIN_STAT == 1:  # solved
                min_oval = cpx.obj.value
            elif MIN_STAT == 2:  # unbound
                min_oval = -numpy.Inf
            elif MIN_STAT == 3:
                # min_oval = pre_sol[R]
                min_oval = numpy.NaN
            else:  # other failure
                min_oval = numpy.NaN
            if debug:
                cpx.write(cpxlp=os.path.join(debug_dir, '%smin.lp' % R))
            if MIN_STAT >= 3:
                if gomin_cntr == 0:
                    cpx.erase()
                    del cpx
                    time.sleep(0.1)
                    cpx = glpk.LPX(mps=mps_filename)
                    gomin_cntr += 1
                else:
                    GOMIN = False
            else:
                GOMIN = False
            if gomin_cntr >= max_error_iter:
                GOMIN = False

        GOMAX = True
        gomax_cntr = 0
        while GOMAX:
            MAX_STAT = 0
            # MAX
            glpk_setObjective(cpx, 'max%s' % R, expr=None, sense='max', reset=False)
            ##  cplx_setBounds(c, id, min=None, max=None) # think about this
            MAX_STAT = glpk_Solve(cpx, method=method)
            if MAX_STAT == 'LPS_OPT':
                MAX_STAT = 1
            elif MAX_STAT == 'LPS_UNBND':
                MAX_STAT = 2
            elif MAX_STAT == 'LPS_NOFEAS':
                MAX_STAT = 3
            else:
                MAX_STAT = 4
            if MAX_STAT == 1:  # solved
                max_oval = cpx.obj.value
            elif MAX_STAT == 2:  # unbound
                max_oval = numpy.Inf
            elif MAX_STAT == 3:  # infeasable
                # max_oval = pre_sol[R]
                max_oval = numpy.NaN
            else:  # other fail
                max_oval = numpy.NaN

            if MAX_STAT >= 3:
                if gomax_cntr == 0:
                    cpx.erase()
                    del cpx
                    time.sleep(0.1)
                    cpx = glpk.LPX(mps=mps_filename)
                    gomax_cntr += 1
                else:
                    GOMAX = False
            else:
                GOMAX = False
            if gomax_cntr >= max_error_iter:
                GOMAX = False

        # check for solver going berserk
        if MIN_STAT > 1 and MAX_STAT > 1:
            print(Ridx)
            time.sleep(1)

        # enables using the default value as a solution if the solver fails
        if pre_opt and default_on_fail:
            if MAX_STAT > 1 and not MIN_STAT > 1:
                max_oval = pre_sol[R]
            if MIN_STAT > 1 and not MAX_STAT > 1:
                min_oval = pre_sol[R]

        if debug:
            cpx.write(cpxlp=os.path.join(debug_dir, '%smax.lp' % R))

        OUTPUT_ARRAY[Ridx, 0] = pre_sol[R]
        if R in REDUCED_COSTS:
            OUTPUT_ARRAY[Ridx, 1] = REDUCED_COSTS[R]
        OUTPUT_ARRAY[Ridx, 2] = min_oval
        OUTPUT_ARRAY[Ridx, 3] = max_oval
        OUTPUT_ARRAY[Ridx, 4] = round(abs(max_oval - min_oval), roundoff_span)
        OUTPUT_ARRAY[Ridx, 5] = MIN_STAT
        OUTPUT_ARRAY[Ridx, 6] = MAX_STAT
        if markupmodel:
            REAC = fba.getReaction(R)
            REAC.setValue(pre_sol[R])
            REAC.fva_min = min_oval
            REAC.fva_max = max_oval
            REAC.fva_status = (MIN_STAT, MAX_STAT)
            if R in REDUCED_COSTS:
                REAC.reduced_costs = REDUCED_COSTS[R]
        if not quiet and MAX_STAT > 1 or MIN_STAT > 1:
            print(
                'Solver fail for reaction \"{}\" (MIN_STAT: {} MAX_STAT: {})'.format(
                    R, MIN_STAT, MAX_STAT
                )
            )
        cntr += 1
        if cntr == 200:
            tcnt += cntr
            print('FVA has processed {} of {} reactions'.format(tcnt, NUM_FLX))
            cntr = 0

    os.remove(mps_filename)

    # cpx.write(cpxlp='thefinaldebug.lp')
    del cpx
    print(
        '\nSinglecore FVA took: {} min (1 process)\n'.format(
            (time.time() - s2time) / 60.0
        )
    )
    print('Output array has columns:')
    print(
        'Reaction, Reduced Costs, Variability Min, Variability Max, abs(Max-Min), MinStatus, MaxStatus'
    )
    return OUTPUT_ARRAY, OUTPUT_NAMES


def glpk_func_GetCPXandPresolve(
    fba,
    pre_opt,
    objF2constr,
    quiet=False,
    oldlpgen=True,
    with_reduced_costs='unscaled',
    method='s',
):
    """
    This is a utility function that does a presolve for FVA, MSAF etc. Generates properly formatted
    empty objects if pre_opt == False

     - *pre_opt* a boolean
     - *fba* a CBModel object
     - *objF2constr* add objective function as constraint
     - *quiet* [default=False] supress cplex output
     - *with_reduced_costs* [default='unscaled'] can be 'scaled' or 'unscaled'
     - *method* [default='s'] select the GLPK solver method, see the GLPK documentation for details

       - 's': simplex
       - 'i': interior
       - 'e': exact


    Returns: pre_sol, pre_oid, pre_oval, OPTIMAL_PRESOLUTION, REDUCED_COSTS

    """

    cpx = glpk_constructLPfromFBA(fba, fname=None)
    OPTIMAL_PRESOLUTION = None
    pre_sol = pre_oid = pre_oval = None
    REDUCED_COSTS = {}
    if pre_opt:
        status = glpk_Solve(cpx, method=method)
        if glpk_getSolutionStatus(cpx) == 'LPS_OPT':
            print('Valid Presolution')
            OPTIMAL_PRESOLUTION = True
            pre_sol, pre_oid, pre_oval = glpk_getOptimalSolution(cpx)
            (
                fba.objectives[fba.activeObjIdx].solution,
                fba.objectives[fba.activeObjIdx].value,
            ) = (pre_sol, pre_oval)
            scaled = False
            if with_reduced_costs == 'scaled':
                REDUCED_COSTS = glpk_getReducedCosts(cpx, scaled=True)
            elif with_reduced_costs == 'unscaled':
                REDUCED_COSTS = glpk_getReducedCosts(cpx, scaled=False)
        else:
            print('Invalid Presolution, because {}'.format(glpk_getSolutionStatus(cpx)))
            OPTIMAL_PRESOLUTION = False
            pre_sol = {}
            for r in fba.reactions:
                pre_sol.update({r.getId(): 0.0})
                r.reduced_cost = 0.0
            pre_oval = 0.0
            pre_oid = 'None'
            raise RuntimeError(
                '\nPresolve failed to optimize this model and cannot continue!'
            )
    else:
        pre_sol = {}
        for r in fba.reactions:
            pre_sol.update({r.getId(): 0.0})
        if objF2constr:
            pre_oval = fba.objectives[fba.activeObjIdx].value
            pre_oid = fba.objectives[fba.activeObjIdx].getId()
        else:
            pre_oval = 0.0
            pre_oid = 'None'
    for r in fba.reactions:
        r.value = pre_sol[r.id]
    return cpx, pre_sol, pre_oid, pre_oval, OPTIMAL_PRESOLUTION, REDUCED_COSTS


def glpk_func_SetObjectiveFunctionAsConstraint(
    cpx, rhs_sense, oval, tol, optPercentage
):
    """
    Take the objective function and "optimum" value and add it as a constraint
     - *cpx* a cplex object
     - *oval* the objective value
     - *tol*  [default=None] do not floor/ceiling the objective function constraint, otherwise round of to *tol*
     - *rhs_sense* [default='lower'] means objC >= objVal the inequality to use for the objective constraint can also be *upper* or *equal*
     - *optPercentage* [default=100.0] means the percentage optimal value to use for the RHS of the objective constraint: optimal_value*(optPercentage/100.0)

    """

    # generate new constraint from old objective value (use non-zero coefficients)
    LCS = []
    LCN = []
    new_constraint = []
    for v_ in range(len(cpx.cols)):
        LCN.append(cpx.cols[v_].name)
        LCS.append(cpx.obj[v_])
        if LCS[-1] != 0.0:
            new_constraint.append((LCS[-1], LCN[-1]))

    if rhs_sense == 'equal':
        pass
    elif cpx.obj.maximize and (rhs_sense == 'upper'):
        print(
            '\nWarning: RHS sense error: \"upper\" does not match \"maximize\" changing to \"lower\"'
        )
        rhs_sense = 'lower'
        time.sleep(1)
    elif not cpx.obj.maximize and (rhs_sense == 'lower'):
        print(
            '\nWarning: RHS sense error: \"lower\" does not match \"minimize\" changing to \"upper\"'
        )
        rhs_sense = 'upper'
        time.sleep(1)
    else:
        print('\nRHS sense ok.')

    # set objective constraint
    if rhs_sense == 'equal':
        cplx_setSingleConstraint(
            cpx, 'ObjCstr', expr=new_constraint, sense='E', rhs=oval
        )
        ##  cplx_setSingleConstraint(cpx, 'ObjCstr', expr=[(1, pre_oid)], sense='E', rhs=oval)
    elif rhs_sense == 'upper':
        if tol != None:
            ub = numpy.ceil(oval / tol) * tol * optPercentage / 100.0
        else:
            ub = oval * (optPercentage / 100.0)
        glpk_setSingleConstraint(cpx, 'ObjCstr', expr=new_constraint, sense='L', rhs=ub)
        ##  cplx_setSingleConstraint(cpx, 'ObjCstr', expr=[(1, pre_oid)], sense='L', rhs=ub)
    elif rhs_sense == 'lower':
        if tol != None:
            lb = numpy.floor(oval / tol) * tol * optPercentage / 100.0
        else:
            lb = oval * (optPercentage / 100.0)
        glpk_setSingleConstraint(cpx, 'ObjCstr', expr=new_constraint, sense='G', rhs=lb)
        ##  cplx_setSingleConstraint(cpx, 'ObjCstr', expr=[(1, pre_oid)], sense='G', rhs=lb)
    else:
        raise RuntimeError("\nInvalid RHS sense: %s" % rhs_sense)


def glpk_setSingleConstraint(c, cid, expr=[], sense='E', rhs=0.0):
    """
    Sets a new sigle constraint to a GLPK model

     - *c* a GLPK instance
     - *cid* the constraint id
     - *expr* a list of (coefficient, name) pairs
     - *sense* [default='G'] LGE
     - *rhs* [default=0.0] the right hand side

    """

    baseRows = len(c.rows)
    c.rows.add(1)
    ofnames = [a[1] for a in expr]
    ofval = [a[0] for a in expr]

    # name and coefficients
    newCon = []
    for cidx_ in range(len(c.cols)):
        if c.cols[cidx_].name in ofnames:
            newCon.append((cidx_, ofval[ofnames.index(c.cols[cidx_].name)]))
            print(newCon[-1])
        else:
            newCon.append((cidx_, 0.0))
    c.rows[baseRows].name = cid
    c.rows[baseRows].matrix = newCon

    # sense and rhs
    if sense in ['<=', '<', 'L']:
        c.rows[baseRows].bounds = None, rhs
    elif sense in ['>=', '>', 'G']:
        c.rows[baseRows].bounds = rhs, None
    elif sense in ['=', 'E']:
        c.rows[baseRows].bounds = rhs
    else:
        raise RuntimeError('INFO: invalid operator: %s' % sense)
    # c.write(cpxlp='test.setc.lp')


def glpk_setObjective(c, oid, expr=None, sense='maximize', reset=True):
    """
    Set a new objective function note that there is a major memory leak in
    `c.variables.get_names()` whch is used when reset=True. If this is a problem
    use cplx_setObjective2 which takes *names* as an input:

     - *c* a GLPK LP object
     - *oid* the r_id of the flux to be optimized
     - *expr* a list of (coefficient, flux) pairs
     - *sense* 'maximize'/'minimize'
     - *reset* [default=True] reset all objective function coefficients to zero

    """
    sense = sense.lower()
    if sense == 'max':
        sense = 'maximize'
    if sense == 'min':
        sense = 'minimize'
    if sense in ['maximise', 'minimise']:
        sense = sense.replace('se', 'ze')
    assert sense in ['maximize', 'minimize'], (
        "\nsense must be ['maximize', 'minimize'] not %s" % sense
    )

    if expr != None:
        ofnames = [a[1] for a in expr]
        ofval = [a[0] for a in expr]

        for cidx_ in range(len(c.cols)):
            if c.cols[cidx_].name in ofnames:
                c.obj[cidx_] = ofval[ofnames.index(c.cols[cidx_].name)]
            elif reset:
                c.obj[cidx_] = 0.0
            else:
                c.obj[cidx_] = c.obj[cidx_]
    c.obj.name = oid

    if sense == 'minimize':
        c.obj.maximize = False
        if __DEBUG__:
            print('Set minimizing')
    else:
        c.obj.maximize = True
        if __DEBUG__:
            print('Set maximizing')
    # c.write(cpxlp='test_obja.lp')


def getReducedCosts(fba):
    """
    Get a dictionary of reduced costs for each reaction/flux

    """
    output = {}
    for r in fba.reactions:
        output.update({r.getId(): r.reduced_cost})
    return output


def setReducedCosts(fba, reduced_costs):
    """
    For each reaction/flux, sets the attribute "reduced_cost" from a dictionary of
    reduced costs

     - *fba* an fba object
     - *reduced_costs* a dictionary of {reaction : value} pairs

    """
    if len(reduced_costs) == 0:
        pass
    else:
        for r in fba.reactions:
            if r.getId() in reduced_costs:
                r.reduced_cost = reduced_costs[r.getId()]
            else:
                r.reduced_cost = None


# def cplx_getDualValues(c):
# """
# Get the get the dual values of the solution

# - *c* a CPLEX LP

# Output is a dictionary of {name : value} pairs

# """
# d_names = c.linear_constraints.get_names()
# d_values = c.solution.get_dual_values()
# output = {}
# for j in range(len(d_names)):
# output.update({d_names[j] : d_values[j]})
# return output

# def cplx_getSensitivities(c):
# """
# Get the sensitivities of each constraint on the objective function with input

# - *c* a CPLEX LP

# Output is a tuple of bound and objective sensitivities where the objective
# sensitivity is described in the CPLEX reference manual as::

# ... the objective sensitivity shows each variable, its reduced cost and the range over
# which its objective function coefficient can vary without forcing a change
# in the optimal basis. The current value of each objective coefficient is
# also displayed for reference.

# - *objective coefficient sensitivity* {flux : (reduced_cost, lower_obj_sensitivity, coeff_value, upper_obj_sensitivity)}
# - *rhs sensitivity* {constraint : (low, value, high)}
# - *bound sensitivity ranges* {flux : (lb_low, lb_high, ub_low, ub_high)}

# """
# SENSE_RHS = {}
# SENSE_BND = {}
# SENSE_OBJ = {}
# c_names = c.linear_constraints.get_names()
# rhs_val = c.linear_constraints.get_rhs()
# j_names = c.variables.get_names()

# rhs_sense = c.solution.sensitivity.rhs()
# bnd_sense = c.solution.sensitivity.bounds()
# obj_sense = c.solution.sensitivity.objective()
# obj_coeff = c.objective.get_linear()
# red_cost = c.solution.get_reduced_costs()

# for r in range(c.variables.get_num()):
# SENSE_BND.update({j_names[r] : (bnd_sense[r][0], bnd_sense[r][1], bnd_sense[r][2], bnd_sense[r][3])})
# SENSE_OBJ.update({j_names[r] : (red_cost[r], obj_sense[r][0], obj_coeff[r], obj_sense[r][1])})

# for s in range(c.linear_constraints.get_num()):
# SENSE_RHS.update({c_names[s] : (rhs_sense[s][0], rhs_val[s], rhs_sense[s][1])})

# return (SENSE_OBJ, SENSE_RHS, SENSE_BND)


def glpk_MinimizeSumOfAbsFluxes(
    fba,
    selected_reactions=None,
    pre_opt=True,
    tol=None,
    objF2constr=True,
    rhs_sense='lower',
    optPercentage=100.0,
    work_dir=None,
    quiet=False,
    debug=False,
    objective_coefficients={},
    return_lp_obj=False,
    oldlpgen=False,
    with_reduced_costs=None,
    method='s',
):
    """
    Minimize the sum of absolute fluxes sum(abs(J1) + abs(J2) + abs(J3) ... abs(Jn)) by adding two constraints per flux
    and a variable representing the absolute value:

     Min: Ci abs_Ji
      Ji - abs_Ji <= 0
      Ji + abs_Ji >= 0

     Such that:
      NJi = 0
      Jopt = opt

    returns the value of the flux minimization objective function (not the model objective function which remains unchanged from)

    Arguments:

     - *fba* an FBA model object
     - *selected reactions* [default=None] means use all reactions otherwise use the reactions listed here
     - *pre_opt* [default=True] attempt to presolve the FBA and report its results in the ouput, if this is disabled and *objF2constr* is True then the vid/value of the current active objective is used
     - *tol*  [default=None] do not floor/ceiling the objective function constraint, otherwise round of to *tol*
     - *rhs_sense* [default='lower'] means objC >= objVal the inequality to use for the objective constraint can also be *upper* or *equal*
     - *optPercentage* [default=100.0] means the percentage optimal value to use for the RHS of the objective constraint: optimal_value*(optPercentage/100.0)
     - *work_dir* [default=None] the MSAF working directory for temporary files default = cwd+fva
     - *debug* [default=False] if True write out all the intermediate MSAF LP's into work_dir
     - *quiet* [default=False] if enabled supress CPLEX output
     - *objF2constr* [default=True] add the model objective function as a constraint using rhs_sense etc. If
       this is True with pre_opt=False then the id/value of the active objective is used to form the constraint
     - *objective_coefficients* [default={}] a dictionary of (reaction_id : float) pairs that provide the are introduced as objective coefficients to the absolute flux value. Note that the default value of the coefficient (non-specified) is +1.
     - *return_lp_obj* [default=False] off by default when enabled it returns the CPLEX LP object
     - *with_reduced_costs* [default=None] if not None should be 'scaled' or 'unscaled'
     - *method* [default='s'] select the GLPK solver method, see the GLPK documentation for details

       - 's': simplex
       - 'i': interior
       - 'e': exact


    With outputs:

     - *fba* an update instance of a CBModel. Note that the FBA model objective function value is the original value set as a constraint

    """

    if with_reduced_costs == 'scaled':
        fba.SCALED_REDUCED_COSTS = True
    elif with_reduced_costs == 'unscaled':
        fba.SCALED_REDUCED_COSTS = False
    elif fba.SCALED_REDUCED_COSTS == True:
        with_reduced_costs = 'scaled'
    elif fba.SCALED_REDUCED_COSTS == False:
        with_reduced_costs = 'unscaled'
    else:
        with_reduced_costs = None

    print('\nINFO: using \"{}\" reduced costs.\n'.format(with_reduced_costs))

    if work_dir == None:
        work_dir = os.getcwd()
    else:
        assert os.path.exists(work_dir), '\nWhat did you think would happen now!'
    if debug:
        debug_dir = os.path.join(work_dir, 'DEBUG')
        if not os.path.exists(debug_dir):
            os.mkdir(debug_dir)

    # generate a presolution
    print('GLPK is using solver option: "{}"'.format(method))
    cpx = OPTIMAL_PRESOLUTION = None
    pre_sol = pre_oid = pre_oval = None
    REDUCED_COSTS = {}
    (
        cpx,
        pre_sol,
        pre_oid,
        pre_oval,
        OPTIMAL_PRESOLUTION,
        REDUCED_COSTS,
    ) = glpk_func_GetCPXandPresolve(
        fba,
        pre_opt,
        objF2constr,
        quiet=quiet,
        oldlpgen=oldlpgen,
        with_reduced_costs=with_reduced_costs,
        method=method,
    )
    # if required add the objective function as a constraint
    if objF2constr:
        glpk_func_SetObjectiveFunctionAsConstraint(
            cpx, rhs_sense, pre_oval, tol, optPercentage
        )

    STORED_OPT = None
    if pre_opt:
        STORED_OPT = pre_oval
    else:
        STORED_OPT = fba.getActiveObjective().getValue()

    # minimize the absolute sum of fluxes
    if selected_reactions != None:
        rids = fba.getReactionIds()
        for r in selected_reactions:
            assert r in rids, "\n%s is not a valid reaction name" % r
    else:
        selected_reactions = fba.getReactionIds()

    # this removes the objective function
    ##  fba_obj_ids = fba.getActiveObjective().getFluxObjectiveReactions()
    ##  print fba_obj_ids
    ##  for R in fba_obj_ids:
    ##  if R in selected_reactions:
    ##  selected_reactions.pop(selected_reactions.index(R))
    ##  del R, fba_obj_ids
    NUM_FLX = len(selected_reactions)
    print('Total number of reactions: {}'.format(NUM_FLX))
    abs_selected_reactions = ['abs_%s' % r for r in selected_reactions]
    # absJ implicitly assumed to be absJ >= 0
    # C#cpx.variables.add(names=abs_selected_reactions)
    basecols = len(cpx.cols)
    cpx.cols.add(len(abs_selected_reactions))
    for v_ in range(len(abs_selected_reactions)):
        cpx.cols[basecols + v_].name = abs_selected_reactions[v_]
        cpx.cols[basecols + v_].bounds = 0, None

    """
     J - abs_J <= 0
     J + abs_J >= 0
    """
    baserows = len(cpx.rows)
    # tnew = time.time()
    rowNames = tuple([c.name for c in cpx.rows])
    colNames = tuple([c.name for c in cpx.cols])
    cRange = range(baserows, baserows + (len(selected_reactions)))
    rcntr = 0
    cpx.rows.add(len(selected_reactions) * 2)
    for r_ in cRange:
        # name and coefficients
        cpx.rows[r_].name = 'abs{}a'.format(selected_reactions[rcntr])
        cpx.rows[r_].matrix = [
            (colNames.index(selected_reactions[rcntr]), 1),
            (colNames.index(abs_selected_reactions[rcntr]), -1),
        ]
        cpx.rows[r_].bounds = None, 0
        cpx.rows[r_ + len(selected_reactions)].name = 'abs{}b'.format(
            selected_reactions[rcntr]
        )
        cpx.rows[r_ + len(selected_reactions)].matrix = [
            (colNames.index(selected_reactions[rcntr]), 1),
            (colNames.index(abs_selected_reactions[rcntr]), 1),
        ]
        cpx.rows[r_ + len(selected_reactions)].bounds = 0, None
        rcntr += 1

    glpk_setObjective(
        cpx, 'MAFS', [(1, j) for j in abs_selected_reactions], 'min', reset=True
    )

    if debug:
        cpx.write(cpxlp=os.path.join(debug_dir, 'MSAF_base_(%s).lp' % time.time()))
        # cplx_writeLPtoLPTfile(cpx, , title=None, Dir=debug_dir)
        ##  cplx_writeLPtoLPTfile(cpx, 'MSAF_base_%s' % time.time() , title=None, Dir=debug_dir)

    glpk_Solve(cpx, method=method)
    glpk_setFBAsolutionToModel(fba, cpx, with_reduced_costs=with_reduced_costs)
    glpk_setSolutionStatusToModel(fba, cpx)

    # cpx.write(cpxlp='test.lp'); return cpx

    minSum = cpx.obj.value
    fba.setAnnotation('min_flux_sum', minSum)
    fba.getActiveObjective().setValue(STORED_OPT)
    print('\nMinimizeSumOfAbsFluxes objective value: {}\n'.format(minSum))
    if quiet:
        pass
        # cplx_setOutputStreams(cpx, mode='default')
    if not return_lp_obj:
        return minSum
    else:
        return cpx


def glpk_writeLPtoLPTfile(c, filename, title=None, Dir=None):
    """
    Write out a GLPK model as an LP format file

    """
    if Dir != None:
        filename = os.path.join(Dir, filename)
    # if title != None:
    # c.set_problem_name(title)
    # c.write(filename+'.lp', filetype='lp')
    c.write(cpxlp=filename + '.lp')
    print('LP output as {}'.format(filename + '.lp'))
